'''
Controller class for the differnetial evolution class diffev
Takes care of stopping and starting - output to the gui as well
as some input from dialog boxes.
'''
from dataclasses import dataclass
from logging import debug

import numpy as np
import time
from typing import Union, TYPE_CHECKING
from threading import Thread, Event

import wx.lib.newevent

from .exception_handling import CatchModelError
from .history_dialog import HistoryDialog
from .message_dialogs import ShowErrorDialog, ShowQuestionDialog, ShowWarningDialog
from .settings_dialog import SettingsDialog
from .custom_events import *
from .. import diffev, fom_funcs, model_control, levenberg_marquardt
from ..remote import optimizer as remote_optimizer
from ..core.custom_logging import iprint
from ..core.colors import COLOR_CYCLES
from ..model_actions import ModelInfluence, ModelAction
from ..plugins.utils import ShowInfoDialog
from ..solver_basis import SolverParameterInfo, SolverResultInfo, SolverUpdateInfo, GenxOptimizerCallback


if TYPE_CHECKING:
    from . import main_window


class GuiCallbacks(GenxOptimizerCallback):
    echo = True

    def __init__(self, parent: wx.Window):
        self.parent = parent

    def text_output(self, text):
        '''
        Function to present the output from the optimizer to the user.
        Takes a string as input.
        '''
        if self.echo:
            debug(f'User Info: {text}', stacklevel=3)
        evt = update_text(text=text)
        wx.QueueEvent(self.parent, evt)

    def plot_output(self, update_data):
        '''
        Solver to present the graphical output from the optimizer to the
        user. Takes the solver as input argument and picks out the
        variables to show in the GUI.
        '''
        evt = update_plot(data=update_data.data, fom_value=update_data.fom_value,
                          fom_name=update_data.fom_name,
                          fom_log=update_data.fom_log, update_fit=update_data.new_best,
                          desc='Fitting update')
        wx.QueueEvent(self.parent, evt)

    def parameter_output(self, param_info):
        '''
        Function to send an update event to update windows that displays
        the parameters to update the values.
        Takes the solver as input argument and picks out the variables to
        show in the GUI.
        '''
        evt = update_parameters(values=param_info.values,
                                new_best=param_info.new_best,
                                population=param_info.population,
                                max_val=param_info.max_val,
                                min_val=param_info.min_val,
                                fitting=True,
                                desc='Parameter Update', update_errors=False,
                                permanent_change=False)
        wx.QueueEvent(self.parent, evt)

    def fitting_ended(self, result_data):
        '''
        function used to post an event when the fitting has ended.
        This must be done since it is not thread safe otherwise. Same GUI in
        two threads when dialogs are run. dangerous...
        '''
        evt = fitting_ended(start_guess=result_data.start_guess,
                            error_message=result_data.error_message,
                            values=result_data.values,
                            new_best=result_data.new_best,
                            population=result_data.population,
                            max_val=result_data.max_val,
                            min_val=result_data.min_val,
                            fitting=True, desc='Fitting Ended')
        wx.QueueEvent(self.parent, evt)

    def autosave(self):
        '''
        Function that conducts an autosave of the model.
        '''
        evt = autosave()
        wx.QueueEvent(self.parent, evt)


class DelayedCallbacks(Thread, GuiCallbacks):
    last_text: Union[list, None] = None
    last_param: Union[SolverParameterInfo, None] = None
    last_update: Union[SolverUpdateInfo, None] = None
    last_endet: Union[SolverResultInfo, None] = None
    min_time = 0.5
    last_iter: float = 0.0
    wait_lock: Event
    stop_thread: Event

    def __init__(self, parent: wx.Window):
        GuiCallbacks.__init__(self, parent)
        self.echo = False
        Thread.__init__(self, daemon=True, name="GenxDelayedCallbacks")
        self.wait_lock = Event()
        self.stop_thread = Event()

    def run(self):
        self.last_iter = time.time()
        self.stop_thread.clear()
        while not self.stop_thread.is_set():
            # main loop for checking for updates and sending GUI events
            time.sleep(max(0., (self.last_iter-time.time()+self.min_time)))
            if self.last_text:
                GuiCallbacks.text_output(self, self.last_text)
                self.last_text = None
            if self.last_param:
                GuiCallbacks.parameter_output(self, self.last_param)
                self.last_param = None
            if self.last_update:
                GuiCallbacks.plot_output(self, self.last_update)
                self.last_update = None
            if self.last_endet:
                GuiCallbacks.fitting_ended(self, self.last_endet)
                self.last_endet = None
            self.last_iter = time.time()
            self.wait_lock.clear()
            self.wait_lock.wait()

    def exit(self):
        self.stop_thread.set()
        self.wait_lock.set()
        self.join(timeout=1.0)

    def text_output(self, text):
        debug(f'User Info: {text}', stacklevel=3)
        self.last_text = text
        self.wait_lock.set()

    def fitting_ended(self, result_data):
        self.last_endet = result_data
        self.wait_lock.set()

    def parameter_output(self, param_info):
        self.last_param = param_info
        self.wait_lock.set()

    def plot_output(self, update_data):
        self.last_update = update_data
        self.wait_lock.set()


@dataclass
class BatchOptions:
    keep_last: bool = True
    adjust_bounds: bool = False


class ModelControlGUI(wx.EvtHandler):
    '''
    Class to take care of the GUI - solver interaction.
    Implements dialogboxes for setting parameters and controls
    for the solver routine. This is where the application specific
    code are used i.e. interfacing the optimization code to the GUI.
    '''
    batch_running: bool = False
    batch_options: BatchOptions

    def __init__(self, parent: 'main_window.GenxMainWindow'):
        wx.EvtHandler.__init__(self)
        self.parent = parent
        self.solvers = {
            'Differential Evolution': diffev.DiffEv(),
            'Levenberg-Marquardt': levenberg_marquardt.LMOptimizer()
            }
        try:
            from ..bumps_optimizer import BumpsOptimizer
        except ImportError:
            pass
        else:
            self.solvers['Bumps'] = BumpsOptimizer()
        self.solvers['Remote DiffEv'] = remote_optimizer.RemoteOptimizer()

        self.controller = model_control.ModelController(self.solvers['Differential Evolution'])
        self.callback_controller = DelayedCallbacks(parent)
        self.callback_controller.start()
        self.controller.set_callbacks(self.callback_controller)
        self.controller.set_action_callback(self.OnActionCallback)
        self.parent.Bind(EVT_FITTING_ENDED, self.OnFittingEnded)
        self.parent.Bind(EVT_AUTOSAVE, self.AutoSave)
        self.batch_options = BatchOptions()

        # Now load the default configuration
        self.ReadConfig()

    def OnActionCallback(self, action: ModelAction):
        self.SetUndoRedoLabels()
        if ModelInfluence.SCRIPT in action.influences:
            self.parent.set_script_text(self.get_model_script())
            evt = update_script(new_script=self.get_model_script())
            wx.PostEvent(self, evt)
        if ModelInfluence.DATA in action.influences:
            cs = self.controller.get_color_cycle()
            colors2keys = dict((value, key) for key, value in COLOR_CYCLES.items())
            if cs in colors2keys:
                self.parent.mb_checkables[colors2keys[cs]].Check()
            else:
                self.parent.mb_checkables[colors2keys[None]].Check()
            dl = self.parent.data_list.list_ctrl
            dl.data_cont.data = self.get_data()
            dl._UpdateImageList()
            dl._UpdateData('Plot settings changed', data_changed=True)
        if ModelInfluence.PARAM in action.influences:
            self.parent.paramter_grid.table.SetParameters(self.controller.get_parameters(),
                                                          clear=False, permanent_change=True)
            evt = value_change()
            wx.PostEvent(self.parent, evt)

    def OnUndo(self, event):
        self.controller.undo_action()

    def OnRedo(self, event):
        self.controller.redo_action()

    def SetUndoRedoLabels(self):
        undos, redos = self.controller.history_stacks()
        if len(undos)!=0:
            self.parent.undo_menu.Enable(True)
            self.parent.undo_menu.SetItemLabel(f"Undo ({undos[-1].action_name})\tCtrl+Z")
        else:
            self.parent.undo_menu.Enable(False)
            self.parent.undo_menu.SetItemLabel("Undo\tCtrl+Z")
        if len(redos)!=0:
            self.parent.redo_menu.Enable(True)
            self.parent.redo_menu.SetItemLabel(f"Redo ({redos[-1].action_name})\tCtrl+Shift+Z")
        else:
            self.parent.redo_menu.Enable(False)
            self.parent.redo_menu.SetItemLabel("Redo\tCtrl+Shift+Z")

    def new_model(self):
        self.controller.new_model()

    def get_model(self):
        return self.controller.get_model()

    def set_model_script(self, text):
        self.controller.set_model_script(text)

    def set_model_params(self, params):
        self.controller.set_model_params(params)

    def get_model_params(self):
        return self.controller.get_model_params()

    def get_model_script(self):
        return self.controller.get_model_script()

    def set_data(self, data):
        self.controller.set_data(data)

    def get_data(self):
        return self.controller.get_data()

    @skips_event
    def update_plotsettings(self, event):
        self.controller.set_data_plotsettings(event.indices, event.sim_par, event.data_par)

    def update_color_cycle(self, source):
        self.controller.update_color_cycle(source)

    def get_parameters(self):
        return self.controller.get_parameters()

    def get_sim_pars(self):
        return self.controller.get_sim_pars()

    def get_parameter_data(self, row):
        return self.controller.get_parameter_data(row)

    def get_parameter_name(self, row):
        return self.controller.get_parameter_name(row)

    def get_possible_parameters(self):
        return self.controller.get_possible_parameters()

    def get_fom(self):
        return self.controller.get_fom()

    def get_fom_name(self):
        return self.controller.get_fom_name()

    def set_filename(self, filename):
        self.controller.set_filename(filename)

    def get_filename(self):
        return self.controller.get_filename()

    def get_model_name(self):
        return self.controller.get_model_name()

    def force_compile(self):
        self.controller.force_compile()

    def compile_if_needed(self):
        self.controller.compile_if_needed()

    def simulate(self, recompile=False):
        self.controller.simulate(recompile=recompile)

    def set_error_pars(self, error_values):
        self.controller.set_error_pars(error_values)

    def export_data(self, basename):
        self.controller.export_data(basename)

    def export_table(self, basename):
        self.controller.export_table(basename)

    def export_script(self, basename):
        self.controller.export_script(basename)

    def export_orso(self, basename):
        self.controller.export_orso(basename)

    def import_table(self, filename):
        self.controller.import_table(filename)

    def import_script(self, filename):
        self.controller.import_script(filename)

    def get_data_as_asciitable(self, indices=None):
        return self.controller.get_data_as_asciitable(indices=indices)

    def set_update_min_time(self, new_time):
        self.callback_controller.min_time = new_time

    @property
    def saved(self):
        return self.controller.saved

    @saved.setter
    def saved(self, value):
        self.controller.saved = value

    @property
    def eval_in_model(self):
        return self.controller.eval_in_model

    @property
    def script_module(self):
        return self.controller.script_module

    def get_solvers(self):
        return list(self.solvers.keys())

    def set_solver(self, name):
        self.controller.optimizer = self.solvers[name]
        self.controller.set_callbacks(self.callback_controller)

    def ReadConfig(self):
        '''
        Reads the parameter that should be read from the config file.
        And set the parameters in both the optimizer and this class.
        '''
        self.controller.ReadConfig()

    def WriteConfig(self):
        '''
        Writes the current configuration of the solver to file.
        '''
        self.controller.WriteConfig()

    def ParametersDialog(self, frame):
        '''
        Shows the Parameters dialog box to set the parameters for the solver.
        '''
        # Update the configuration if a model has been loaded after
        # the object have been created..
        self.ReadConfig()
        fom_func_name = self.controller.get_fom_name()
        if not fom_func_name in fom_funcs.func_names:
            ShowWarningDialog(self.parent, 'The loaded fom function, ' \
                              +fom_func_name+', does not exist '+ \
                              'in the local fom_funcs file. The fom fucntion has been'+
                              ' temporary added to the list of availabe fom functions')
            fom_funcs.func_names.append(fom_func_name)
            exectext = 'fom_funcs.'+fom_func_name+ \
                       ' = self.parent.model.fom_func'
            exec(exectext, locals(), globals())

        combined_options = self.controller.get_combined_options()
        dlg = SettingsDialog(frame, combined_options,
                             apply_callback=lambda options: False,
                             title='Optimizer Settings')

        res = dlg.ShowModal()
        if res==wx.ID_OK:
            updates = dlg.collect_results()
            self.controller.update_combined_options(updates)
        dlg.Destroy()

    def ModelLoaded(self):
        '''
        Function that takes care of resetting everything when a model has
        been loaded.
        '''
        evt = update_plot(model=self.controller.get_fitted_model(),
                          fom_log=self.controller.get_fom_log(), update_fit=False,
                          desc='Model loaded')
        wx.PostEvent(self.parent, evt)
        self.controller.history_clear()

        # Update the parameter plot ...
        if self.controller.is_configured():
            # remember to add a check
            res = self.controller.get_result_info()
            try:
                evt = update_parameters(values=res.values,
                                        new_best=False,
                                        population=res.population,
                                        max_val=res.par_max,
                                        min_val=res.par_min,
                                        fitting=True,
                                        desc='Parameter Loaded', update_errors=False,
                                        permanent_change=False)
            except AttributeError:
                iprint('Could not create data for parameters')
            else:
                wx.PostEvent(self.parent, evt)
        wx.CallLater(100, self.SetModelSaved)

    def SetModelSaved(self):
        self.saved = True

    @skips_event
    def OnFittingEnded(self, evt):
        '''
        Callback when fitting has ended. Takes care of cleaning up after
        the fit. Calculates errors on the parameters and updates the grid.
        '''
        if evt.error_message:
            ShowErrorDialog(self.parent, evt.error_message)
            return

        if not self.batch_running:
            message = 'Do you want to keep the parameter values from the fit?'
            result = ShowQuestionDialog(self.parent, message, 'Keep the fit?', yes_no=True)
        else:
            result = True
        if result:
            self.controller.set_value_pars(evt.values)
            evt = update_parameters(values=evt.values,
                                    new_best=True,
                                    population=evt.population,
                                    max_val=evt.max_val,
                                    min_val=evt.min_val,
                                    fitting=False,
                                    desc='Parameter Improved', update_errors=False,
                                    permanent_change=True)
            wx.PostEvent(self.parent, evt)
            if self.batch_running:
                wx.CallAfter(self.batch_next)
        else:
            evt = update_parameters(values=evt.start_guess,
                                    new_best=True,
                                    population=evt.population,
                                    max_val=evt.max_val,
                                    min_val=evt.min_val,
                                    fitting=False,
                                    desc='Parameter Reset', update_errors=False,
                                    permanent_change=False)
            wx.PostEvent(self.parent, evt)

    def batch_next(self):
        # move to the next dataset in the batch and start fitting
        idx = self.controller.active_index()
        params = self.get_parameters()
        if idx+1==len(self.controller.model_store):
            self.batch_running = False
            evt = batch_next(last_index=idx, finished=True)
            wx.QueueEvent(self, evt)
            return
        self.controller.activate_model(idx+1)
        new_pars = self.get_parameters()
        if self.batch_options.keep_last:
            row_nmb, funcs, values, min_, max_ = params.get_fit_pars()
            for ri, vi in zip(row_nmb, values):
                new_pars.set_value(ri, 1, vi)
        if self.batch_options.adjust_bounds:
            row_nmb, funcs, values, min_, max_ = new_pars.get_fit_pars()
            for ri, vi, mii, mai in zip(row_nmb, values, min_, max_):
                val_range = (mai-mii)
                new_pars.set_value(ri, 3, vi-val_range/2.)
                new_pars.set_value(ri, 4, vi+val_range/2.)
        # wait 1 second for other GUI updates to finish before starting next fit
        evt=batch_next(last_index=idx, finished=False)
        wx.QueueEvent(self, evt)
        wx.CallLater(1000, self.controller.StartFit)

    @skips_event
    def OnSetParameterValue(self, evt):
        self.controller.set_parameter_value(evt.row, evt.col, evt.value)

    @skips_event
    def OnMoveParameter(self, evt):
        self.controller.move_parameter(evt.row, evt.step)

    @skips_event
    def OnInsertParameter(self, evt):
        self.controller.insert_parameter(evt.row)

    @skips_event
    def OnDeleteParameter(self, evt):
        self.controller.delete_parameter(evt.rows)

    @skips_event
    def OnSortAndGroupParameters(self, evt):
        self.controller.sort_and_group_parameters(evt.sort_params)

    @skips_event
    def OnUpdateParameters(self, evt):
        if evt.desc not in ['Parameter Update', 'Parameter Reset']:
            return
        self.parent.paramter_grid.table.ShowParameters(evt.values)

    def OnShowHistory(self, evt):
        dia = HistoryDialog(self.parent, self.controller.history)
        res = dia.ShowModal()
        if dia.changed_actions:
            self.OnActionCallback(dia.changed_actions)
        dia.Destroy()

    def CalcErrorBars(self):
        res = self.controller.CalcErrorBars()
        if (res[:, 0]>0.).any() or (res[:, 1]<0.).any():
            ShowInfoDialog(self.parent,
                           "There is something wrong in the error estimation, low/high values don't have the right sign.\n\n"
                           "This can be caused by non single-modal parameter statistics, closeness to bounds or too low value of"
                           "'burn' before stampling.",
                           title="Issue in uncertainty estimation")
        error_strings = []
        for error_low, error_high in res:
            error_str = '(%.3e, %.3e)'%(error_low, error_high)
            error_strings.append(error_str)
        return error_strings

    def ProjectEvals(self, parameter):
        return self.controller.ProjectEvals(parameter)

    def ScanParameter(self, parameter, points):
        '''
        Scans one parameter and records its fom value as a function 
        of the parameter value.
        '''
        row = parameter
        model = self.controller.model
        (funcs, vals) = model.get_sim_pars()
        minval = model.parameters.get_data()[row][3]
        maxval = model.parameters.get_data()[row][4]
        parfunc = funcs[model.parameters.get_sim_pos_from_row(row)]
        par_def_val = vals[model.parameters.get_sim_pos_from_row(row)]
        step = (maxval-minval)/points
        par_vals = np.arange(minval, maxval+step, step)
        fom_vals = np.array([])

        par_name = model.parameters.get_data()[row][0]
        dlg = wx.ProgressDialog("Scan Parameter",
                                "Scanning parameter "+par_name,
                                maximum=len(par_vals),
                                parent=self.parent,
                                style=wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME
                                      | wx.PD_REMAINING_TIME | wx.PD_AUTO_HIDE)
        with CatchModelError(self.parent, 'ScanParameter', 'scan through values') as cme:
            # Start with setting all values
            [f(v) for (f, v) in zip(funcs, vals)]
            for par_val in par_vals:
                parfunc(par_val)
                fom_vals = np.append(fom_vals, model.evaluate_fit_func())
                dlg.Update(len(fom_vals))
        dlg.Destroy()
        # resetting the scanned parameter
        parfunc(par_def_val)
        if cme.successful:
            return par_vals, fom_vals

    def ResetOptimizer(self):
        pass

    def StartFit(self):
        self.controller.StartFit()

    def StopFit(self):
        if self.batch_running:
            # if batch processing, finish last fit if not pushed a second time
            self.batch_running = False
            return
        self.controller.StopFit()

    def ResumeFit(self):
        self.controller.ResumeFit()

    def IsFitted(self):
        return self.controller.IsFitted()

    @skips_event
    def AutoSave(self, _event):
        self.controller.save()

    def load_file(self, fname):
        prog = wx.ProgressDialog('Loading...', f'Reading from file\n{fname}\n',
                                 maximum=100, parent=self.parent)

        def update_callback(i, N):
            prog.Update(int(i/N*100), f'Reading from file\n{fname}\ndataset {i} of {N}')

        try:
            self.controller.load_file(fname, update_callback=update_callback)
        finally:
            prog.Destroy()
        solver_classes = [si.__class__ for si in self.solvers.values()]
        loaded_solver = self.controller.optimizer.__class__
        if loaded_solver in solver_classes:
            current_solver = list(self.solvers.keys())[solver_classes.index(loaded_solver)]
        else:
            self.solvers[loaded_solver.__name__] = self.controller.optimizer
            current_solver = loaded_solver.__name__
            self.parent.eh_ex_add_solver_selection(current_solver)
        self.parent.eh_ex_set_solver_selection(current_solver)

    def set_error_bars_level(self, value):
        '''
        Sets the value of increase of the fom used for errorbar calculations
        '''
        if value<1:
            raise ValueError('fom_error_bars_level has to be above 1')
        else:
            self.controller.optimizer.opt.errorbar_level = value

    def set_save_all_evals(self, value):
        '''
        Sets the boolean value to save all evals to file
        '''
        self.controller.optimizer.opt.save_all_evals = bool(value)



# Functions for instrumental resolution corrections and area of illumination
# these are to be included in the Models class and Refl class
# Note that these function assumes a Gaussian spread in both the angular as well
# as in real space

# THis one really needs scipy
from numpy import (pi, sqrt, sin, cos, tan, arcsin, arccos, arctan, arctan2, real, imag, abs, cumsum, zeros, isscalar,
                   where, ones, linspace, exp, newaxis, transpose, arange, convolve, r_, array, not_equal, logical_not,
                   sum, dot, minimum, maximum, float32, float64, complex128, ndarray, float32, complex64, c_, r_)
try:
    # trapz removed in numpy 2.x
    from numpy import trapezoid as trapz
except ImportError:
    from numpy import trapz
from scipy.special import erf
from genx.core.custom_logging import iprint


rad = pi/180.
sqrt2 = sqrt(2.)


# Intensity correction Gaussian beamprofile
def GaussIntensity(alpha, s1, s2, sigma_x):
    sinalpha = sin(alpha*rad)
    if s1==s2:
        return erf(s2/sqrt2/sigma_x*sinalpha)
    else:
        common = sinalpha/sqrt2/sigma_x
        return (erf(s2*common)+erf(s1*common))/2.0


# Specular foorprintcorrections square beamprofile
def SquareIntensity(alpha, slen, beamwidth):
    F = slen/beamwidth*sin(alpha*rad)
    return where(F<=1.0, F, ones(F.shape))


####################################################################
## Resolution Functions (Normal Distributed=Gaussian)
#####################################################################

# Full Convlutions - vayring resolution

# Function to create a 1D vector for the resolution with the 
# positions to calculate the reflectivity Qret and the weight
# of each point weight
# Inputs: Q - the Q values 
#         dQ - the resolution
#         points - the number of points for the convolution
#         range how far the gaussian should be convoluted
def ResolutionVector(Q, dQ, points, range=3):
    # if type(dQ)!=type(array([])):
    #    dQ=dQ*ones(Q.shape)
    Qres = Q+dQ*linspace(-range, range, points)[:, newaxis]

    weight = 1/sqrt(2*pi)/dQ*exp(-(transpose(Q[:, newaxis])-Qres)**2/dQ**2/2)
    Qret = Qres.flatten()  # reshape(Qres,(1,Qres.shape[0]*Qres.shape[1]))[0]
    # print Qres
    # print Qres.shape
    # print Qret.shape
    return Qret, weight


# Include the resolution with Qret and weight calculated from ResolutionVector
# and I the calculated intensity at each point. returns the intensity
def ConvoluteResolutionVector(Qret, I, weight):
    Qret2 = Qret.reshape(weight.shape[0], weight.shape[1])
    # print Qret.shape,weight.shape
    I2 = I.reshape(weight.shape[0], weight.shape[1])
    # print (I*weight).shape,Qret.shape
    norm_fact = trapz(weight, x=Qret2, axis=0)
    Int = trapz(I2*weight, x=Qret2, axis=0)/norm_fact
    # print Int.shape
    return Int


# Fast convlution - constant resolution
# constant spacing between data!
def ConvoluteFast(Q, I, dQ, range=3):
    Qstep = Q[1]-Q[0]
    resvector = arange(-range*dQ, range*dQ+Qstep, Qstep)
    weight = 1/sqrt(2*pi)/dQ*exp(-resvector**2/dQ**2/2)
    Iconv = convolve(r_[ones(resvector.shape)*I[0], I, ones(resvector.shape)*I[-1]], weight/weight.sum(),
                     mode='same')[resvector.shape[0]:-resvector.shape[0]]
    return Iconv


# Fast convolution - varying resolution
# constant spacing between the dat.
def ConvoluteFastVar(Q, I, dQ, range=3):
    Qstep = Q[1]-Q[0]
    steps = max(dQ*ones(Q.shape))*range/Qstep
    weight = 1/sqrt(2*pi)/dQ*exp(-(Q[:, newaxis]-Q)**2/dQ**2/2)
    Itemp = I[:, newaxis]*ones(I.shape)
    norm_fact = trapz(weight, axis=0)
    Int = trapz(Itemp*weight, axis=0)/norm_fact
    return Int


def QtoTheta(wavelength, Q):
    return arcsin(wavelength/4.0/pi*Q)/rad


def TwoThetatoQ(wavelength, TwoTheta):
    return 4.0*pi/wavelength*sin(rad/2.*TwoTheta)


########## Neutron Polarization #################
def get_pol_matrix(p1, p2, F1, F2):
    # @formatter:off
    P1 = array([
                [       1,       0,       0,      0],
                [       0,       1,       0,      0],
                [      F1,       0, (1.-F1),      0],
                [       0,      F1,       0, (1.-F1)],
               ])
    P2 = array([
                [       1,       0,       0,      0],
                [      F2, (1.-F2),       0,      0],
                [       0,       0,       1,      0],
                [       0,       0,      F2, (1.-F2)],
               ])
    P3 = array([
                [ (1.-p1),       0,      p1,       0],
                [       0, (1.-p1),       0,      p1],
                [      p1,       0, (1.-p1),       0],
                [       0,      p1,       0, (1.-p1)],
               ])
    P4 = array([
                [ (1.-p2),      p2,       0,       0],
                [      p2, (1.-p2),       0,       0],
                [       0,       0, (1.-p2),      p2],
                [       0,       0,      p2, (1.-p2)],
               ])
    # @formatter:on
    P = dot(dot(P1, P2), dot(P3, P4))
    return P


from . import USE_NUMBA


if USE_NUMBA:
    # try to use numba to speed up the calculation intensive functions:
    try:
        from .instrument_numba import GaussIntensity, SquareIntensity, QtoTheta, TwoThetatoQ, ResolutionVector
    except Exception as e:
        iprint('Could not use numba, no speed up from JIT compiler:\n'+str(e))



# -*- coding: utf8 -*-
'''
=============
LayerGraphics
=============

A plugin to create sketch graphics to vizualize the layer structure.
The result is an SVG image with automatically generated colors and sizes.

Thicknesses can either be linear or mapped through a scaling funciton to
avoid large variation if thicknesses are too different.
'''

import svgwrite
import wx

from wx.svg import SVGimage
from dataclasses import dataclass
from numpy import log, array, sqrt, pi
from genx.models.lib.refl import InstrumentBase, LayerBase, SampleBase, StackBase
from .. import add_on_framework as framework

try:
    # noinspection PyUnresolvedReferences
    from typing import List
except ImportError:
    List = object

@dataclass
class LColor:
    r: float
    g: float
    b: float

    def __add__(self, other):
        out = LColor(self.r+other.r, self.g+other.g, self.b+other.b)

    def normalize(self):
        norm = sqrt((array([self.r, self.g, self.b])**2).sum())
        self.r/=norm
        self.g/=norm
        self.b/=norm

    @property
    def side(self):
        # return lighter color for side of 3d block
        return LColor(min(1.0, self.r+0.35), min(1.0, self.g+0.35), min(1.0, self.b+0.35))

    @property
    def top(self):
        # return lighter color for side of 3d block
        return LColor(min(1.0, self.r+0.2), min(1.0, self.g+0.2), min(1.0, self.b+0.2))

    def __str__(self):
        return f"rgb({self.r*255:.0f}, {self.g*255:.0f}, {self.b*255:.0f})"

    def get_wx(self):
        return wx.Colour(int(self.r*225), int(self.g*255), int(self.b*255))

@dataclass
class LInfo:
    thickness: float
    scale: float
    color: LColor

COLORS = [LColor(1., 0., 0.), LColor(0., 1., 0.), LColor(0., 0., 1.),
          LColor(1., 1., 0.), LColor(1., 0., 1.), LColor(0., 1., 1.)]

class BlockGenerator:
    stacks: List[StackBase]
    rescale: bool  # apply the mapping function to reduce visible variation of size
    show_all: bool # show all layers in a repeating stack separately
    show_one: bool # show just one sequence in a repreating stack when above 1 repetition

    dmin: float
    dmax: float
    dtotal: float

    repetitions: List[int]
    thicknesses: List[List[float]]
    scaled: List[List[float]]

    min_rescale = 30.
    max_rescale = 100.

    def __init__(self, stacks: List[StackBase], rescale=True, show_all=False, show_one=False):
        self.stacks = [si for si in stacks if len(si.Layers)>0]
        self.rescale = rescale
        self.show_all = show_all
        self.show_one = show_one
        self.get_sample_range()

    def rescale_thickness(self, di):
        ldi = log(di)
        ldmin = log(self.dmin)
        ldmax = log(self.dmax)
        if ldmin==ldmax:
            # don't resace if all thicknesses are the same
            return di
        ld = (ldi-ldmin)/(ldmax-ldmin)
        sd = self.min_rescale + (self.max_rescale-self.min_rescale)*ld
        return sd

    def get_sample_range(self):
        # return total thickness, smallest and largest layer
        dmin = 1.0e5
        dmax = 0.
        dtotal = 0.

        repetitions = []
        thicknesses = []
        color_slds = []

        for stack in self.stacks:
            repetitions.append(int(stack.Repetitions))
            thicknesses.append([li.d for li in stack.Layers])
            color_slds.append([abs(li.dens*li.b)+abs(li.dens*li.f) for li in stack.Layers])
            if len(thicknesses[-1])>0:
                dmin = min(dmin, min(thicknesses[-1]))
                dmax = max(dmax, max(thicknesses[-1]))

        unique_ids={}
        color_ids=[]
        cid = 0
        for group_slds in color_slds:
            group_ids=[]
            for csld in group_slds:
                csld = round(csld, 4)
                if csld not in unique_ids:
                    unique_ids[csld]=cid
                    cid+=1
                group_ids.append(unique_ids[csld])
            color_ids.append(group_ids)

        self.dmin=dmin
        self.dmax=dmax
        self.repetitions = repetitions
        self.color_ids = color_ids

        if self.rescale:
            sthicknesses = [list(map(self.rescale_thickness, ti)) for ti in thicknesses]
        else:
            sthicknesses = thicknesses
        self.thicknesses = thicknesses
        self.scaled = sthicknesses

        for ri, ti in zip(repetitions, sthicknesses):
            dsequence=sum(ti)
            if ri<2:
                dtotal+=dsequence
            elif self.show_one:
                dtotal += dsequence + dsequence / 2
            elif self.show_all or ri==2:
                dtotal += ri*dsequence
            else:
                dtotal += 2*dsequence + dsequence / 2
        self.dtotal = dtotal

    def get_blocks(self):
        """
        Returns a list of blocks with either a thickness or list of thicknesses
        """
        output = []
        for ri, ti, si, cidi in zip(self.repetitions, self.thicknesses, self.scaled, self.color_ids):
            infos = [LInfo(ti_j, si_j, COLORS[cid_j%len(COLORS)]) for ti_j, si_j, cid_j in zip(ti, si, cidi)]
            dsequence = sum(si)
            if ri<2:
                output += infos
            elif self.show_one:
                output.append(infos+[LInfo(-ri, dsequence/2, LColor(1., 1., 1.))])
            elif self.show_all or ri==2:
                output.append(ri*infos)
            else:
                output.append(infos+[LInfo(-ri, dsequence/2, LColor(1., 1., 1.))]+infos)
        return output

    def __repr__(self):
        blocks = self.get_blocks()
        bstr = 'substrate |'
        sstr = ' |'
        for bi in blocks:
            if isinstance(bi, list):
                bstr+=f'|'
                sstr+=f'|'
                for bij in bi:
                    sstr+=f' {bij.scale:.0f} |'
                    if bij.thickness<0:
                        bstr=bstr[:-1]
                        bstr+=' (x%i) '%(-bij.thickness)
                    else:
                        bstr+=f' {bij.thickness:.0f} |'
                bstr += f'|'
                sstr += f'|'
            else:
                bstr+=f' {bi.thickness:.0f} |'
                sstr += f' {bi.scale:.0f} |'
        bstr += ' ambient'
        return f'BlockGenerator(\n               {bstr}\n               {sstr})'

class SVGenerator:
    block_generator: BlockGenerator
    svg: svgwrite.Drawing
    use3d: bool

    unit_precision: int # number of digits behind decimal point for thickness label
    unit_nm: bool # use nm instead of angstrom as unit

    view_box = array([100, 200])
    move_3d = 8.0
    vanishing_point=(175., -15.)
    fontsize = 10

    def __init__(self, sample: SampleBase, rescale=True, show_all=False, show_one=False, use3d=True,
                 unit_precision=1, unit_nm=True, fontsize=10):
        self.use3d = use3d
        self.fontsize = fontsize
        self.unit_precision = unit_precision
        self.unit_nm = unit_nm
        self.block_generator = BlockGenerator(sample.Stacks, rescale=rescale, show_all=show_all, show_one=show_one)
        self.create_svg()

    def to_vp(self, point0):
        """
        Calculate the x,y-values for a shift from a starting point (x0,y0)
        towards the vanishing point by an x-offset of self.move_3d.
        """
        px, py=self.view_box*0.01
        x0, y0 = point0
        dx = self.move_3d
        vx, vy = self.vanishing_point
        relx = dx/(vx-x0) # releative movement towrads vanishing point
        dy = (vy-y0)*relx
        return ((x0+dx)*px, (y0+dy)*py)

    def tlabel(self, thickness):
        if self.unit_nm:
            return f"{thickness/10:.{self.unit_precision}f} nm"
        else:
            return f"{thickness:.{self.unit_precision}f} Å"

    def add_rect(self, pos:float, si:float, color:LColor=LColor(0., 0., 0.), in_stack=False,
                 width=90):
        px, py=self.view_box*0.01
        if in_stack:
            x= 5 + self.fontsize
            w = width-self.fontsize
        else:
            x = 5
            w = width
        if self.use3d:
            w -= 10
            points = [((x+w)*px, pos*py),
                      ((x+w)*px, (pos+si)*py),
                      self.to_vp(((x+w), (pos+si))),
                      self.to_vp(((x+w), pos))]
            side = self.svg.polygon(points=points,
                                    stroke='black', stroke_width=0.5,
                                    fill=str(color.side))
            self.svg.add(side)
        rect = self.svg.rect(
            (x*px, pos*py),
            (w*px, si*py),
            stroke='black', stroke_width=0.5,
            fill=str(color))
        self.svg.add(rect)

    def create_svg(self):
        px, py = self.view_box*0.01
        bg = self.block_generator
        self.svg = svgwrite.Drawing('genx_model.svg', size=("4cm", "8cm"),
                                    viewBox=f'0 0 {self.view_box[0]} {self.view_box[1]}')
        vscale = bg.dtotal / 85. # 90% of image height to be used for layers

        blocks = bg.get_blocks()

        if self.use3d:
            tc = 45
        else:
            tc = 50

        pos = 90.0
        self.add_rect(pos, 100, LColor(0, 0, 0))
        bc = LColor(0,0,0)
        for bi in blocks:
            if isinstance(bi, list):
                block_length = sum([bij.scale/vscale for bij in bi])
                self.add_rect(pos-block_length, block_length, LColor(0.75, 0.75, 0.75), width=10+self.fontsize)
                for bij in bi:
                    dij = bij.scale/vscale
                    pos -= dij
                    if bij.thickness<0:
                        if self.use3d:
                            # add surface polygon
                            shifted = self.to_vp((85, pos+dij))
                            points = [((5+self.fontsize)*px, (pos+dij)*py),
                                      (85*px, (pos+dij)*py),
                                      shifted,
                                      (shifted[0]-(85-5-self.fontsize)*px, shifted[1])]
                            side = self.svg.polygon(points=points,
                                                    stroke='black', stroke_width=0.5,
                                                    fill=str(bc.top))
                            self.svg.add(side)
                        paragraph = self.svg.add(self.svg.g(font_size=self.fontsize))
                        paragraph.add(self.svg.text(" ... ", (tc*px, (pos+dij/3.)*py),
                                                    text_anchor='middle', dominant_baseline='middle'))
                        paragraph = self.svg.add(self.svg.g(font_size=self.fontsize,
                                                            transform=f"translate(5, {(pos+dij)*py}) rotate(-90)"))
                        paragraph.add(self.svg.text(f"(x%i)"%(-bij.thickness), (0, 0),
                                                    text_anchor='middle', dominant_baseline='hanging'))
                    else:
                        bc = bij.color
                        self.add_rect(pos, dij, bc, in_stack=True)
                        paragraph = self.svg.add(self.svg.g(font_size=self.fontsize))
                        paragraph.add(self.svg.text(self.tlabel(bij.thickness), (tc*px, (pos+dij/2.)*py),
                                                    text_anchor='middle', dominant_baseline='middle'))
            else:
                di = bi.scale/vscale
                pos -= di
                bc = bi.color
                self.add_rect(pos, di, bc)
                paragraph = self.svg.add(self.svg.g(font_size=self.fontsize))
                paragraph.add(self.svg.text(self.tlabel(bi.thickness), (tc*px, (pos+di/2.)*py),
                                            text_anchor='middle', dominant_baseline='middle'))
        if self.use3d:
            # add surface polygon
            shifted = self.to_vp((85, pos))
            points = [(5*px, pos*py),
                      (85*px, pos*py),
                      shifted,
                      (shifted[0]-80*px, shifted[1])]
            side = self.svg.polygon(points=points,
                                    stroke='black', stroke_width=0.5,
                                    fill=str(bc.top))
            self.svg.add(side)

    def add_rect_gc(self, gc:wx.GraphicsContext, pos:float, si:float, color:LColor=LColor(0., 0., 0.), in_stack=False,
                 width=90):
        px, py=self.view_box*0.01
        gx, gy = gc.GetSize()
        gc_scale = min(gx/self.view_box[0], gy/self.view_box[1])

        if in_stack:
            x= 5 + self.fontsize
            w = width-self.fontsize
        else:
            x = 5
            w = width
        if self.use3d:
            w -= 10
            points = [((x+w)*px, pos*py),
                      ((x+w)*px, (pos+si)*py),
                      self.to_vp(((x+w), (pos+si))),
                      self.to_vp(((x+w), pos))]
            self.draw_polygon_gc(gc, points, color.side)
        gc.SetPen(wx.BLACK_PEN)
        gc.SetBrush(wx.Brush(color.get_wx(), style=wx.BRUSHSTYLE_SOLID))
        gc.DrawRectangle(x*px*gc_scale, pos*py*gc_scale, w*px*gc_scale, si*py*gc_scale)

    def draw_polygon_gc(self, gc:wx.GraphicsContext, points, fill=LColor(1, 1, 1)):
        gx, gy = gc.GetSize()
        gc_scale = min(gx/self.view_box[0], gy/self.view_box[1])

        gc.SetPen(wx.BLACK_PEN)
        gc.SetBrush(wx.Brush(fill.get_wx(), style=wx.BRUSHSTYLE_SOLID))

        points2d=[wx.Point2D(gc_scale*x, gc_scale*y) for x,y in points]
        points2d.append(wx.Point2D(gc_scale*points[0][0], gc_scale*points[0][1]))

        gc.DrawLines(points2d)


    def draw_centered_text(self, gc:wx.GraphicsContext, txt:str, x, y, baseline=False):
        w, h = gc.GetFullTextExtent(txt)[:2]
        if baseline:
            gc.DrawText(txt, int(x-w/2.), int(y))
        else:
            gc.DrawText(txt, int(x-w/2.), int(y-h/2.))

    def render_to_gc(self, gc:wx.GraphicsContext):
        px, py = self.view_box*0.01
        bg = self.block_generator
        gx, gy = gc.GetSize()
        gc_scale = min(gx/self.view_box[0], gy/self.view_box[1])

        vscale = bg.dtotal / 85. # 90% of image height to be used for layers

        blocks = bg.get_blocks()

        gc.SetFont(wx.Font(int(self.fontsize*gc_scale*0.75),
                           wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL), wx.BLACK)

        if self.use3d:
            tc = 45
        else:
            tc = 50

        pos = 90.0
        self.add_rect_gc(gc, pos, 100, LColor(0, 0, 0))
        bc = LColor(0,0,0)
        for bi in blocks:
            if isinstance(bi, list):
                block_length = sum([bij.scale/vscale for bij in bi])
                self.add_rect_gc(gc, pos-block_length, block_length, LColor(0.75, 0.75, 0.75), width=10+self.fontsize)
                for bij in bi:
                    dij = bij.scale/vscale
                    pos -= dij
                    if bij.thickness<0:
                        if self.use3d:
                            # add surface polygon
                            shifted = self.to_vp((85, pos+dij))
                            points = [((5+self.fontsize)*px, (pos+dij)*py),
                                      (85*px, (pos+dij)*py),
                                      shifted,
                                      (shifted[0]-(85-5-self.fontsize)*px, shifted[1])]
                            self.draw_polygon_gc(gc, points, bc)
                        self.draw_centered_text(gc, "...", gc_scale*tc*px, gc_scale*(pos+dij/3.)*py)
                        gc.DrawText(f"(x%i)"%(-bij.thickness), gc_scale*5, gc_scale*(pos+dij)*py, angle=pi/2)
                    else:
                        bc = bij.color
                        self.add_rect_gc(gc, pos, dij, bc, in_stack=True)
                        self.draw_centered_text(gc, self.tlabel(bij.thickness),
                                                gc_scale*tc*px, gc_scale*(pos+dij/2.)*py)
            else:
                di = bi.scale/vscale
                pos -= di
                bc = bi.color
                self.add_rect_gc(gc, pos, di, bc)
                self.draw_centered_text(gc, self.tlabel(bi.thickness),
                                        gc_scale*tc*px, gc_scale*(pos+di/2.)*py)
        if self.use3d:
            # add surface polygon
            shifted = self.to_vp((85, pos))
            points = [(5*px, pos*py),
                      (85*px, pos*py),
                      shifted,
                      (shifted[0]-80*px, shifted[1])]
            self.draw_polygon_gc(gc, points, bc)


class SVGPanel(wx.Panel):
    #svg_img: SVGimage = None
    svg_img: SVGenerator = None
    last_scale = 1.0

    def __init__(self, parent):
        super().__init__(parent=parent, style=wx.FULL_REPAINT_ON_RESIZE)
        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def OnPaint(self, event: wx.PaintEvent):
        if self.svg_img is not None:
            img = self.svg_img

            dc = wx.PaintDC(self)
            dc.SetBackground(wx.Brush('white'))
            dc.Clear()

            ctx = wx.GraphicsContext.Create(dc)
            img.render_to_gc(ctx)
            #img.RenderToGC(ctx, scale)
        else:
            event.Skip()


class Plugin(framework.Template):
    svg: str = None

    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent=parent

        # Create the SLD plot
        LG_panel=self.NewPlotFolder('Layer Graphics')
        SA_sizer=wx.BoxSizer(wx.VERTICAL)
        LG_panel.SetSizer(SA_sizer)

        self.img = SVGPanel(LG_panel)
        SA_sizer.Add(self.img, 1, wx.EXPAND | wx.GROW | wx.ALL)

        bot_sizer=wx.BoxSizer(wx.HORIZONTAL)
        SA_sizer.Add(bot_sizer, 0, wx.FIXED_MINSIZE)
        left_sizer=wx.BoxSizer(wx.VERTICAL)
        bot_sizer.Add(left_sizer, 1, wx.EXPAND | wx.GROW | wx.FIXED_MINSIZE)
        bot_sizer.AddSpacer(4)
        mid_sizer=wx.BoxSizer(wx.VERTICAL)
        bot_sizer.Add(mid_sizer, 1, wx.EXPAND | wx.GROW | wx.FIXED_MINSIZE)
        bot_sizer.AddSpacer(4)
        right_sizer=wx.BoxSizer(wx.VERTICAL)
        bot_sizer.Add(right_sizer, 1, wx.EXPAND | wx.GROW | wx.FIXED_MINSIZE)


        self.rescale = wx.CheckBox(LG_panel, label='Rescale thickness display')
        self.rescale.SetValue(True)
        left_sizer.Add(self.rescale, 0, wx.FIXED_MINSIZE)

        self.show_all = wx.RadioButton(LG_panel, label='All layers')
        left_sizer.Add(self.show_all, 0, wx.FIXED_MINSIZE)

        show_topbot = wx.RadioButton(LG_panel, label='Top/Bottom layers')
        show_topbot.SetValue(True)
        left_sizer.Add(show_topbot, 0, wx.FIXED_MINSIZE)

        self.show_one = wx.RadioButton(LG_panel, label='Single Repetition')
        left_sizer.Add(self.show_one, 0, wx.FIXED_MINSIZE)

        self.use3d = wx.CheckBox(LG_panel, label='Pseudo 3d')
        self.use3d.SetValue(True)
        mid_sizer.Add(self.use3d, 0, wx.FIXED_MINSIZE)
        self.unit_nm = wx.CheckBox(LG_panel, label='nm-unit')
        self.unit_nm.SetValue(True)
        mid_sizer.Add(self.unit_nm, 0, wx.FIXED_MINSIZE)
        self.unit_precision = wx.SpinCtrl(LG_panel, value="1", min=0, max=5)
        mid_sizer.Add(self.unit_precision, 0, wx.FIXED_MINSIZE)

        self.fontsize = wx.SpinCtrl(LG_panel, value="10", min=1, max=20)
        right_sizer.Add(self.fontsize, 0, wx.FIXED_MINSIZE)
        export_button = wx.Button(LG_panel, label='Save to SVG...')
        right_sizer.Add(export_button)
        copy_button = wx.Button(LG_panel, label='Copy to Clipboard')
        right_sizer.Add(copy_button)

        LG_panel.Layout()
        try:
            self.OnSimulate(None)
        except Exception:
            pass

        self.rescale.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.show_all.Bind(wx.EVT_RADIOBUTTON, self.OnSimulate)
        show_topbot.Bind(wx.EVT_RADIOBUTTON, self.OnSimulate)
        self.show_one.Bind(wx.EVT_RADIOBUTTON, self.OnSimulate)
        self.use3d.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.unit_nm.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.unit_precision.Bind(wx.EVT_SPINCTRL, self.OnSimulate)
        self.fontsize.Bind(wx.EVT_SPINCTRL, self.OnSimulate)
        export_button.Bind(wx.EVT_BUTTON, self.ExportSVG)
        copy_button.Bind(wx.EVT_BUTTON, self.CopyImage)

    def OnSimulate(self, event):
        # Calculate and update the sld plot
        model = self.GetModel()
        if model.script_module is None:
            return
        gen=SVGenerator(model.script_module.sample,
                        rescale=self.rescale.GetValue(),
                        show_all=self.show_all.GetValue(),
                        show_one=self.show_one.GetValue(),
                        use3d=self.use3d.GetValue(),
                        unit_nm=self.unit_nm.GetValue(),
                        unit_precision=int(self.unit_precision.GetValue()),
                        fontsize=self.fontsize.GetValue(),
                        )
        self.svg = gen.svg.tostring()
        #self.img.svg_img = SVGimage.CreateFromBytes(self.svg.encode('utf-8'))
        self.img.svg_img = gen
        self.img.Refresh()

    def ExportSVG(self, event):
        dlg = wx.FileDialog(self.img, message="Save layer sketch image", defaultFile="",
                            wildcard="SVG Image|*.svg",
                            style=wx.FD_SAVE | wx.FD_CHANGE_DIR | wx.FD_OVERWRITE_PROMPT
                            )
        if dlg.ShowModal()==wx.ID_OK:
            path = dlg.GetPath()
            with open(path, 'w', encoding='utf-8') as fh:
                fh.write(self.svg)
        dlg.Destroy()

    def CopyImage(self, event):
        bmp = wx.Bitmap(wx.Size(300, 600), depth=32)
        if hasattr(bmp, 'SetScaleFactor'):
            bmp.SetScaleFactor(2.0)

        memdc = wx.MemoryDC(bmp)
        memdc.SetBackground(wx.Brush(wx.Colour(255, 255, 255, 0)))
        memdc.Clear()

        ctx = wx.GraphicsContext.Create(memdc)

        self.img.svg_img.render_to_gc(ctx)
        memdc.SelectObject(wx.NullBitmap)

        bmp_obj = wx.BitmapDataObject()
        bmp_obj.SetBitmap(bmp)

        if not wx.TheClipboard.IsOpened():
            open_success = wx.TheClipboard.Open()
            if open_success:
                wx.TheClipboard.SetData(bmp_obj)
                wx.TheClipboard.Close()
                wx.TheClipboard.Flush()



"""
Funcitoniality for checking online for new versions, prompting the user and downloading potential updates.
"""

import os
import sys
import requests
import base64
import tempfile
import subprocess
import wx
import wx.html as html

from logging import debug
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from .help import rst_html
from .output_dialog import TextOutputDialog
from ..version import __version__ as GENX_VERSION

GITHUB_URL="https://api.github.com/repos/aglavic/genx/releases/latest"
GITHUB_TAGS="https://api.github.com/repos/aglavic/genx/tags"


def _get_Retry_kw():
    if hasattr(Retry.DEFAULT, "allowed_methods"):
        return {"allowed_methods": ["HEAD", "GET", "OPTIONS"]}
    return {"method_whitelist": ["HEAD", "GET", "OPTIONS"]}


retry_strategy = Retry(
    total=3,
    status_forcelist=[429, 500, 502, 503, 504],
    **_get_Retry_kw()
)


def check_version():
    adapter = HTTPAdapter(max_retries=retry_strategy)
    http = requests.Session()
    http.mount("https://", adapter)

    try:
        response = http.get(GITHUB_URL).json()
    except Exception as e:
        # if not connection to server is possible, silently ignore
        debug('Could not check for update, error in request:', exc_info=True)
        return True
    if not 'name' in response:
        debug(f"Response from GitHub unexpected: {response!r}")
        return True
    version=response["name"]
    return version[1:]==GENX_VERSION

class VersionInfoDialog(wx.Dialog):
    def __init__(self, parent):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER,
                           name=f'GenX update')
        self.SetTitle(f'GenX update available')
        self.SetSize(wx.Size(800, 800))
        info=self.collect_release_info()
        changes=self.filter_readme(info['readme_text'])

        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        vbox.Add(wx.StaticText(self, label=f'Newest Release: {info["version"]}'))
        vbox.Add(wx.StaticText(self, label=f'Date: {info["date"]}'))
        vbox.Add(wx.StaticText(self, label=f'\nChanges since {GENX_VERSION}'))

        self.html_win=html.HtmlWindow(self, -1,
                                      style=wx.NO_FULL_REPAINT_ON_RESIZE)
        vbox.Add(self.html_win, 1, flag=wx.EXPAND, border=20)
        self.html_win.SetPage(rst_html(changes))

        btn_box=wx.BoxSizer(wx.HORIZONTAL)
        vbox.Add(btn_box)
        self.download_link=None
        if sys.platform.startswith('win') and info['setup_file']:
            # assume Windows users have installed from setup
            self.download_link=info['setup_file']
            self.download_size=info['setup_file_size']
            btn=wx.Button(self, label='Download and Install...')
            btn_box.Add(btn)
            self.Bind(wx.EVT_BUTTON, self.download_and_install, btn)
        if self.check_pip_version():
            btn=wx.Button(self, label='Pip in-place update...')
            btn_box.Add(btn)
            self.Bind(wx.EVT_BUTTON, self.update_pip, btn)

    def check_pip_version(self):
        try:
            import pkg_resources
        except ImportError:
            return False
        pkg_names=[si.key for si in pkg_resources.working_set]
        pkg_versions=[si.version for si in pkg_resources.working_set]
        if 'genx3' in pkg_names and pkg_versions[pkg_names.index('genx3')]==GENX_VERSION:
            return True
        return False

    def download_and_install(self, evt):
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        debug(f'Downloading file {self.download_link}')
        dia=wx.ProgressDialog('Downloading GenX...', 'The latest GenX setup is being downloaded from Github.\n'
                                                     'The setup will start once the download is completed.',
                              maximum=100, parent=None, style=wx.PD_APP_MODAL|wx.PD_AUTO_HIDE)
        dia.Show()
        with http.get(self.download_link, allow_redirects=True, stream=True) as res:
            res.raise_for_status()
            with tempfile.NamedTemporaryFile('wb', delete=False, suffix='.exe', prefix='genx_setup_') as tmp:
                debug(f'Save to temporary file f{tmp.name}')
                s=0
                for chunk in res.iter_content(chunk_size=8192):
                    s+=len(chunk)
                    rel_step=s/self.download_size
                    dia.Update(int(rel_step*100))
                    tmp.write(chunk)
        debug(f'Calling setup file f{tmp.name}')
        dia.Destroy()
        subprocess.Popen(tmp.name)
        debug(f'Started process f{tmp.name}')
        self.EndModal(wx.ID_DELETE)

    def update_pip(self, evt):
        dia=TextOutputDialog(self)
        dia.SetSize(wx.Size(800,500))
        dia.Show()
        dia.write('Starting "pip install genx3 --upgrade":\n\n')
        with subprocess.Popen([sys.executable, '-m', 'pip', 'install', 'genx3', '--upgrade'],
                              stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as proc:
            while proc.poll() is None:
                dia.write(proc.stdout.readline().decode('utf-8'))
        dia.write(f'\n\nProcess finished with exit code {proc.poll()}')
        wx.Yield()
        dia.ShowModal()
        dia.Destroy()
        if proc.poll()==0:
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)

    @staticmethod
    def get_file(http, tree, path):
        path_items = path.split('/')
        for pi in path_items[:-1]:
            for ti in tree:
                if ti['path'] == pi:
                    tree = http.get(ti['url']).json()['tree']
                    break
        for ti in tree:
            if ti['path'] == path_items[-1]:
                data=http.get(ti['url']).json()['content']
                return base64.b64decode(data).decode('utf-8')

    @staticmethod
    def collect_release_info():
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)

        response = http.get(GITHUB_URL).json()
        output=dict(date=response['published_at'], version=response['name'][1:],
                    setup_file=None, deb_file=None)
        for asset in response['assets']:
            if asset['name'].endswith('.exe'):
                output['setup_file']=asset['browser_download_url']
                output['setup_file_size']=asset['size']
            elif asset['name'].endswith('py38.deb'):
                output['deb_file']=asset['browser_download_url']
        tags = requests.get(GITHUB_TAGS).json()
        tag = tags[[ti['name'] for ti in tags].index(response['tag_name'])]
        commit = http.get(tag['commit']['url']).json()
        ctmp = http.get(commit['url']).json()
        tree = http.get(ctmp['commit']['tree']['url']).json()
        output['readme_text'] = VersionInfoDialog.get_file(http, tree['tree'], 'genx/README.txt')
        return output

    @staticmethod
    def filter_readme(txt):
        # shorten the readme to only show changes since this version
        start = txt.find("Change")
        end=txt.find(GENX_VERSION)
        txt=txt[start:end]
        return txt.rsplit('\n', 1)[0]



'''data_loader_framework.py

Library that implements a template (Template) class for classes that
loads data into GenX.
'''
from ..data import DataSet, META_DEFAULT
from .utils import ShowInfoDialog


class Template:
    wildcard=None

    def __init__(self, parent=None):
        self.parent=parent

        if parent is not None:
            # This is made for the virtual datalist controller...
            self.data=self.parent.data_cont.get_data()
            self.Register()

    def Register(self):
        '''
        Register the data loader with the parent
        '''
        self.parent.data_loader=self

    def SetData(self, data):
        '''
        Sets the data connection to the plugin.
        '''
        self.data=data

    def UpdateDataList(self):
        '''
        Forces the data list to update. This is only necessary if new
        data sets have been added when the data has been loaded
        '''
        # Just a force update of the data_list
        self.parent.SetItemCount(
            self.parent.data_cont.get_count())
        # Updating the imagelist as well
        self.parent._UpdateImageList()

    def SetStatusText(self, text):
        '''
        Set a status text in the main frame for user information
        '''
        self.parent.SetStatusText(text)

    def CountDatasets(self, file_path):
        '''
        Count the number of dataset in a file. Default implementation of
        dataloader allows for only one item.
        '''
        return 1

    def GetWildcardString(self):
        if self.wildcard is None:
            wc = "All files (*.*)|*.*"
        else:
            wc = "%s (%s)|%s|All files (*.*)|*.*"%(self.__module__.rsplit('.', 1)[1].capitalize(),
                                                   self.wildcard, self.wildcard)
        return wc

    def LoadDataFile(self, selected_items):
        '''
        Selected items is the selcted items in the items in the current DataList
        into which data from file(s) should be loaded. Note that the default
        implementation only allows the loading of a single file! 
        Overriding this function in subclasses can of course change this 
        behaviour. This function calls the LoadData function which implements 
        the io function by it self. The LoadData has to be overloaded in
        order to have a working plugin.
        '''
        import wx
        n_selected=len(selected_items)
        if n_selected>0:
            dlg=wx.FileDialog(self.parent, message="Choose your Datafile"
                              , defaultFile="", wildcard=self.GetWildcardString()
                              , style=wx.FD_OPEN | wx.FD_CHANGE_DIR)

            if dlg.ShowModal()==wx.ID_OK:
                file_path=dlg.GetPath()
                dlg.Destroy()
                dcount=self.CountDatasets(file_path)
                if n_selected>dcount:
                    ShowInfoDialog(self.parent, f'You can only load {dcount} dataset(s) from this file',
                                   'Too many selections')
                    dlg.ShowModal()
                    dlg.Destroy()
                else:
                    self.SetData(self.parent.data_cont.get_data())
                    for di in range(n_selected):
                        dataset=DataSet(copy_from=self.data[selected_items[di]])
                        self.LoadDataset(dataset, file_path, data_id=di)

                        if len(dataset.x)==0:
                            continue

                        self.data[selected_items[di]]=dataset
                    # In case the dataset name has changed
                    self.UpdateDataList()

                    # Send an update that new data has been loaded
                    self.SendUpdateDataEvent()

                    return True
            else:
                dlg.Destroy()
        else:
            ShowInfoDialog(self.parent, 'Please select a dataset', 'No active dataset')
        return False

    def LoadDataset(self, dataset, file_path, data_id=0):
        '''
        A wrapper around LoadData that makes sure there is minimal metadata generated.
        '''
        dataset.meta = META_DEFAULT.copy()
        # in case the data loader does not define any metadata
        # at least set the instrument to data loader name
        dataset.meta['data_source']['experiment']['instrument'] = self.__module__.rsplit('.', 1)[1]
        self.LoadData(dataset, file_path, data_id=data_id)
        dataset.meta['data_source']['file_name'] = file_path

    def LoadData(self, dataset, file_path, data_id=0):
        '''
        This file should load a single data file into data object of
        the model. Please overide this function. It is called by the 
        LoadFile function.
        '''
        pass

    def SettingsDialog(self):
        '''
        This function should - if necessary implement a dialog box
        that allows the user set import settings for example.
        '''
        pass

    def SendUpdateDataEvent(self):
        '''
        Sends an event that new data has been loaded and 
        things such as plotting should be updated.
        '''
        self.parent._UpdateData('New data added')

    def Remove(self):
        '''
        Removes the plugin from knowledge of the parent.
        '''
        self.parent.data_loader=None

    def CanOpen(self, file_path):
        """
        Return if the data loader class can open the given file.
        Default implementation just checks if the filename matches the wildcard, if it exists.
        """
        if self.wildcard is None:
            return True
        for wc in self.wildcard.split(';'):
            if file_path.endswith(wc[1:]):
                return True
        return False



"""
A dialog to import batches of datafiles to create a model sequence and perform
fitting to a common model.
"""
import numpy as np
import wx.grid
from copy import copy, deepcopy

from .solvergui import ModelControlGUI
from .custom_events import EVT_BATCH_NEXT
from .metadata_dialog import MetaDataDialog
from .plot_dialog import PlotDialog
from ..plugins.data_loader_framework import Template as DataLoaderTemplate


def get_value_from_path(data, path):
    if len(path)>1:
        return get_value_from_path(data[path[0]], path[1:])
    else:
        return data[path[0]]

class CopyGrid(wx.grid.Grid):
    """ Grid calss that copies all data on ctrl+C, no matter the selection. """

    def __init__(self, *args, **opts):
        wx.grid.Grid.__init__(self, *args, **opts)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKey)

    def OnKey(self, event):
        # If Ctrl+C is pressed...
        if event.ControlDown() and event.GetKeyCode()==67:
            self.copy()
        else:
            event.Skip()
            return

    def copy(self):
        output = ''
        for col in range(self.GetNumberCols()):
            name = self.GetColLabelValue(col).replace("\n", ".")
            output += f'{name}\t'
        output += '\n'
        for row in range(self.GetNumberRows()):
            for col in range(self.GetNumberCols()):
                value = self.GetCellValue(row, col)
                output += f'{value}\t'
            output += '\n'

        clipboard = wx.TextDataObject()
        clipboard.SetText(output)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(clipboard)
            wx.TheClipboard.Close()
        else:
            wx.MessageBox("Can't open the clipboard", "Error")


class BatchDialog(wx.Dialog):
    """
    A dialog to display all models that are stored in the file, to generate sequences from
    data files and to fit batches to a common model.
    """
    model_control: ModelControlGUI

    def __init__(self, parent, model_control: ModelControlGUI):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self.SetTitle('Model/dataset list for batch fitting')

        self.dpi_scale_factor = wx.GetApp().dpi_scale_factor
        self.model_control = model_control
        self.model_control.Bind(EVT_BATCH_NEXT, self.OnBatchNext)
        self.plots = {}

        vbox.Add(wx.StaticText(self, label='Below is the list of datasets for batch fitting.\n'
                                           'If you want to activate one dataset, double-click the row label on the left.'),
                 proportion=0, flag=wx.FIXED_MINSIZE | wx.ALL, border=4)

        self.grid = CopyGrid(self)
        self.grid.SetDefaultRowSize(int(self.grid.GetDefaultRowSize()*1.5))
        self.grid.SetDefaultCellAlignment(wx.ALIGN_LEFT, wx.ALIGN_CENTRE)
        vbox.Add(self.grid, proportion=1, flag=wx.EXPAND)
        self.build_grid()

        self.grid.Bind(wx.grid.EVT_GRID_CELL_CHANGING, self.OnUpdateCell)
        self.grid.Bind(wx.grid.EVT_GRID_LABEL_LEFT_DCLICK, self.OnLabelDClick)

        self.keep_last = wx.CheckBox(self, label='Keep result values from last')
        self.keep_last.SetValue(self.model_control.batch_options.keep_last)
        vbox.Add(self.keep_last, proportion=0, flag=wx.FIXED_MINSIZE | wx.TOP | wx.LEFT, border=4)

        self.adjust_bounds = wx.CheckBox(self, label='Adjust boundaries around last values')
        self.adjust_bounds.SetValue(self.model_control.batch_options.adjust_bounds)
        vbox.Add(self.adjust_bounds, proportion=0, flag=wx.FIXED_MINSIZE | wx.LEFT, border=4)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        vbox.Add(hbox, proportion=0, flag=wx.EXPAND)

        btn = wx.Button(self, -1, label='Import Data...')
        btn.SetToolTip('Use active dataset as template and load\na list of datasets to generate a batch.')
        hbox.Add(btn, proportion=0, flag=wx.FIXED_MINSIZE)
        self.Bind(wx.EVT_BUTTON, self.OnImportData, btn)

        btn = wx.Button(self, -1, label='Clear list')
        hbox.Add(btn, proportion=0, flag=wx.FIXED_MINSIZE)
        self.Bind(wx.EVT_BUTTON, self.OnClearData, btn)

        hbox.AddSpacer(int(20*self.dpi_scale_factor))

        btn = wx.Button(self, -1, label='Fit All')
        hbox.Add(btn, proportion=0, flag=wx.FIXED_MINSIZE)
        self.Bind(wx.EVT_BUTTON, self.OnBatchFitModel, btn)

        btn = wx.Button(self, -1, label='Fit From Here')
        hbox.Add(btn, proportion=0, flag=wx.FIXED_MINSIZE)
        self.Bind(wx.EVT_BUTTON, self.OnBatchFromHere, btn)

        hbox.AddSpacer(int(20*self.dpi_scale_factor))

        btn = wx.Button(self, -1, label='Sort by Value')
        hbox.Add(btn, proportion=0, flag=wx.FIXED_MINSIZE)
        self.Bind(wx.EVT_BUTTON, self.OnSort, btn)

        btn = wx.Button(self, -1, label='Extract Value')
        btn.SetToolTip('Use metadata from datasets to extract a sequence value (e.g. temperature)')
        hbox.Add(btn, proportion=0, flag=wx.FIXED_MINSIZE)
        self.Bind(wx.EVT_BUTTON, self.OnExtract, btn)

    def OnBatchNext(self, evt):
        # evt.finished
        self.switch_line(evt.last_index, self.model_control.controller.active_index())
        g = self.grid
        prev_model = self.model_control.controller.model_store[evt.last_index]
        prev_models = self.model_control.controller.model_store[:evt.last_index+1]

        row_nmb, funcs, values, min_, max_ = prev_model.parameters.get_fit_pars()
        x = np.array([mi.sequence_value for mi in prev_models])

        for i, (ni, fi, pi) in enumerate(zip(row_nmb, funcs, values)):
            g.SetCellValue(evt.last_index, i+2, f'{pi:.7g}')
            if fi in self.plots:
                vals = np.array([mi.parameters[ni].value for mi in prev_models])
                plot = self.plots[fi]
                plot.clear_data()
                plot.plot(x, vals, marker='o', ms=2, lw=1, color=(0.0, 0.0, 1.0))
                plot.draw()

    def switch_line(self, prev_row, row):
        g = self.grid
        if prev_row>=0:
            g.SetCellBackgroundColour(prev_row, 0, "#ffffff")
            g.SetCellBackgroundColour(prev_row, 1, "#ffffff")
        if row>=0:
            g.SetCellBackgroundColour(row, 0, "#ff8888")
            g.SetCellBackgroundColour(row, 1, "#ff8888")
        g.MakeCellVisible(row, 0)
        g.ForceRefresh()

    def build_grid(self):
        g = self.grid
        models = self.model_control.controller.model_store
        g.CreateGrid(len(models), 2)
        g.SetSelectionMode(wx.grid.Grid.GridSelectNone)

        g.SetColLabelValue(0, 'item')
        g.SetColLabelValue(1, 'index/value')
        self.fill_grid()

    def fill_grid(self):
        g = self.grid
        models = self.model_control.controller.model_store
        ci = self.model_control.controller.active_index()
        for i, mi in enumerate(models):
            g.SetRowLabelValue(i, f'{i}:')
            g.SetCellValue(i, 0, mi.h5group_name)
            g.SetCellValue(i, 1, f'{mi.sequence_value:.7g}')
            g.DisableRowResize(i)
            if i==ci:
                g.SetCellBackgroundColour(i, 0, "#ff8888")
                g.SetCellBackgroundColour(i, 1, "#ff8888")
            else:
                g.SetCellBackgroundColour(i, 0, "#ffffff")
                g.SetCellBackgroundColour(i, 1, "#ffffff")

        g.AutoSizeColumns(False)

    def OnUpdateCell(self, event: wx.grid.GridEvent):
        if event.GetCol()==0:
            names = [m.h5group_name for m in self.model_control.controller.model_store]
            new_name = event.GetString()
            if new_name in names:
                event.Veto()
                return
            # change the name of a dataaset
            self.model_control.controller.model_store[event.GetRow()].h5group_name = new_name
        elif event.GetCol()==1:
            try:
                value = float(event.GetString())
            except ValueError:
                event.Veto()
                return
            else:
                self.model_control.controller.model_store[event.GetRow()].sequence_value = value
        else:
            event.Veto()

    def OnLabelDClick(self, event: wx.grid.GridEvent):
        if event.GetCol()<0:
            ci = self.model_control.controller.active_index()
            row = event.GetRow()
            self.model_control.controller.activate_model(row)
            self.switch_line(ci, row)

    def set_batch_params(self):
        self.model_control.batch_running = True
        self.model_control.batch_options.keep_last = self.keep_last.IsChecked()
        self.model_control.batch_options.adjust_bounds = self.adjust_bounds.IsChecked()
        self.set_batch_columns()

    def set_batch_columns(self):
        g = self.grid
        if g.GetNumberCols()>2:
            g.DeleteCols(2, g.GetNumberCols()-2)
        row_nmb, funcs, values, min_, max_ = self.model_control.get_parameters().get_fit_pars()
        ncols = len(row_nmb)
        g.AppendCols(ncols)
        for i, ci in enumerate(funcs):
            g.SetColLabelValue(i+2, ci)
            self.create_graph(ci)

    def create_graph(self, parameter):
        if parameter in self.plots:
            self.plots[parameter].Show()
            return
        plot = PlotDialog(self, title=f'Batch parameter {parameter} fit')
        self.plots[parameter] = plot
        plot.Show()
        plot.set_xlabel('batch index/value')
        plot.set_ylabel(parameter)
        plot.draw()

    def OnImportData(self, evt):
        origin = self.model_control.get_model().deepcopy()
        # TODO: There needs to be better separation, data_loader should be accessible through ModelControlGUI
        data_loader: DataLoaderTemplate = self.model_control.parent.data_list.list_ctrl.data_loader
        flists = []
        for i in range(len(origin.data)):
            dia = wx.FileDialog(self, message=f"Select datafiles for dataset {i}",
                                wildcard=data_loader.GetWildcardString(),
                                style=wx.FD_OPEN | wx.FD_CHANGE_DIR | wx.FD_MULTIPLE)
            res = dia.ShowModal()
            if res!=wx.ID_OK:
                return
            flists.append(dia.GetPaths())

        N = len(flists[0])
        prog = wx.ProgressDialog('Importing...', f'Read from files...\n0/{N}',
                                 maximum=100, parent=self)

        def update_callback(i):
            prog.Update(int(i/N*100), f'Read from files...\n{i}/{N}')

        try:
            self.model_control.controller.read_sequence(data_loader, flists, name_by_file=True,
                                                        callback=update_callback)
        finally:
            prog.Destroy()

        self.grid.AppendRows(len(self.model_control.controller.model_store)-self.grid.GetNumberRows())
        self.fill_grid()

    def OnClearData(self, evt):
        self.model_control.controller.model_store = []
        self.grid.DeleteRows(0, self.grid.GetNumberRows())

    def OnBatchFitModel(self, evt):
        evt.Skip()
        # synchronizing current model script and parameters to all datasets
        script = self.model_control.get_model_script()
        params = self.model_control.get_model_params()
        for mi in self.model_control.controller.model_store:
            mi.set_script(copy(script))
            mi.parameters = deepcopy(params)
        # set first model and start fitting
        self.switch_line(self.model_control.controller.active_index(), 0)
        self.model_control.controller.activate_model(0)
        self.set_batch_params()
        wx.CallLater(1000, self.model_control.StartFit)

    def OnBatchFromHere(self, evt):
        evt.Skip()
        # synchronizing current model script and parameters to all datasets
        script = self.model_control.get_model_script()
        params = self.model_control.get_model_params()
        ci = self.model_control.controller.active_index()
        for mi in self.model_control.controller.model_store[ci+1:]:
            mi.set_script(copy(script))
            mi.parameters = deepcopy(params)
        # set first model and start fitting
        self.set_batch_params()
        # put values for parameters from previous datasets
        g = self.grid
        fit_rows, *_ = self.model_control.get_parameters().get_fit_pars()
        for i, mi in enumerate(self.model_control.controller.model_store[:ci+1]):
            params = [mi.parameters[fi].value for fi in fit_rows]
            for j, pj in enumerate(params):
                g.SetCellValue(i, j+2, f'{pj:.7g}')

        wx.CallLater(1000, self.model_control.StartFit)

    def OnSort(self, evt):
        sort_list = [(mi.sequence_value, i, mi) for i, mi in enumerate(self.model_control.controller.model_store)]
        sort_list.sort()
        self.model_control.controller.model_store = [si[2] for si in sort_list]
        self.fill_grid()

    def OnExtract(self, evt):
        dia = MetaDataDialog(self, self.model_control.controller.model_store[0].data,
                             filter_leaf_types=[int, float], close_on_activate=True)
        dia.SetTitle('Select key to use by double-click')
        dia.ShowModal()
        dia.Destroy()
        if dia.activated_leaf is None:
            return

        didx = dia.activated_leaf[0]
        dpath = dia.activated_leaf[1:]
        for mi in self.model_control.controller.model_store:
            try:
                value = float(get_value_from_path(mi.data[didx].meta, dpath))
            except Exception:
                value = 0.
            mi.sequence_value = value
        self.fill_grid()



"""
A simple dialog window to display meta data read from files to the user.
"""

import yaml
import wx
from logging import debug
from datetime import datetime
from copy import deepcopy

from genx.data import DataList
from genx.model import Model


class MetaDataDialog(wx.Dialog):
    datasets: DataList

    def __init__(self, parent, datasets: DataList, selected=0, filter_leaf_types=None, close_on_activate=False):
        wx.Dialog.__init__(self, parent, title="Dataset information",
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX | wx.TR_HIDE_ROOT)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(sizer)

        self.tree = wx.TreeCtrl(self)
        self.leaf_ids = []

        vsizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(vsizer, proportion=1, flag=wx.EXPAND)
        vsizer.Add(self.tree, proportion=1, flag=wx.EXPAND)

        btn = wx.Button(self, label='Make ORSO Conform')
        vsizer.Add(btn, flag=wx.FIXED_MINSIZE)
        btn.Bind(wx.EVT_BUTTON, self.make_orso_conform)

        self.text = wx.TextCtrl(self, style=wx.TE_READONLY | wx.TE_MULTILINE | wx.TE_DONTWRAP)
        sizer.Add(self.text, proportion=2, flag=wx.EXPAND)

        self.datasets = datasets
        self.filter_leaf_types = filter_leaf_types
        try:
            from orsopy import fileio
            self.orso_repr = [fileio.Orso(**deepcopy(di.meta)) for di in self.datasets]
        except Exception:
            self.orso_repr = [None for di in self.datasets]
        self.build_tree(selected)

        self.Bind(wx.EVT_TREE_SEL_CHANGED, self.show_item)
        self.tree.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.item_activated)

        self.tree.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, self.OnRightClick)

        self.SetSize((800, 800))
        self.activated_leaf = None
        self.close_on_activate = close_on_activate

    def build_tree(self, selected):
        root = self.tree.AddRoot('datasets')
        self.tree.SetItemData(root, ('', 'Select key to show information'))
        for i, di in enumerate(self.datasets):
            branch = self.tree.AppendItem(root, di.name)
            self.tree.SetItemData(branch, (di.name,
                                           yaml.dump(di.meta, indent=4, sort_keys=False
                                                     ).replace('    ', '\t').replace('\n', '\n\t'),
                                           [i], self.orso_repr[i]))

            self.add_children(branch, di.meta, [i], self.orso_repr[i])
            if i==selected:
                self.tree.Expand(branch)
        self.tree.Expand(root)
        self.add_coloring()

    def add_children(self, node, source, path, orso_repr):
        for key, value in source.items():
            if isinstance(value, dict):
                obj = getattr(orso_repr, key, None)
                itm = self.tree.AppendItem(node, key)
                self.tree.SetItemData(itm,
                                      (key, yaml.dump(value, indent=4, sort_keys=False
                                                      ).replace('    ', '\t').replace('\n', '\n\t'),
                                       path+[key], obj))
                self.add_children(itm, value, path+[key], getattr(orso_repr, key, None))
            else:
                itm = self.tree.AppendItem(node, key)
                vtype=type(value)
                self.tree.SetItemData(itm, (key, f'{value} ({vtype.__name__})', path+[key], vtype))
                if self.filter_leaf_types is None or type(value) in self.filter_leaf_types:
                    self.leaf_ids.append(itm)
                    self.tree.SetItemBackgroundColour(itm, wx.Colour('aaaaff'))
                else:
                    self.tree.SetItemTextColour(itm, wx.Colour('aaaaaa'))
            if orso_repr:
                if key in getattr(orso_repr, '_orso_optionals', {}):
                    self.tree.SetItemBackgroundColour(itm, wx.Colour(255, 255, 150))
                elif key in getattr(orso_repr, 'user_data', {}):
                    self.tree.SetItemBackgroundColour(itm, wx.Colour(255, 150, 255))
                elif key in getattr(orso_repr, '__annotations__', {}):
                    self.tree.SetItemBackgroundColour(itm, wx.Colour(150, 255, 150))

    def add_coloring(self):
        root = self.tree.GetRootItem()
        node = self.tree.AppendItem(root, 'coloring')
        itm = self.tree.AppendItem(node, 'orso required')
        self.tree.SetItemBackgroundColour(itm, wx.Colour(150, 255, 150))
        itm = self.tree.AppendItem(node, 'orso optional')
        self.tree.SetItemBackgroundColour(itm, wx.Colour(255, 255, 150))
        itm = self.tree.AppendItem(node, 'user defined')
        self.tree.SetItemBackgroundColour(itm, wx.Colour(255, 150, 255))
        self.tree.Expand(node)

    def show_item(self, event: wx.TreeEvent):
        item = event.GetItem()
        name, data, *_ = self.tree.GetItemData(item)
        self.text.Clear()
        self.text.AppendText('%s:\n\n\t%s'%(name, data))

    def item_activated(self, event: wx.TreeEvent):
        if event.GetItem() not in self.leaf_ids:
            event.Skip()
            return
        # a leaf item was activated
        name, data, self.activated_leaf, vtype=self.tree.GetItemData(event.GetItem())
        if self.close_on_activate:
            self.EndModal(wx.ID_OK)
        else:
            prev_dict = self.datasets[self.activated_leaf[0]].meta
            if vtype is type(None):
                from orsopy import fileio
                from typing import Dict, List, Tuple, Union, Literal
                item = fileio.Orso(**prev_dict)
                for key in self.activated_leaf[1:-1]:
                    item = getattr(item, key, None)
                    if item is None:
                        break
                if item is None:
                    vtype = None
                else:
                    vtype = item.__annotations__.get(self.activated_leaf[-1], str)
                from orsopy.fileio.base import get_args, get_origin
                if get_origin(vtype) == Union:
                    vtype = get_args(vtype)[0]
                if get_origin(vtype) == Literal or vtype==fileio.Polarization:
                    if vtype==fileio.Polarization:
                        options = [v.value for v in  fileio.Polarization]
                    else:
                        options = get_args(vtype)
                    dia = wx.SingleChoiceDialog(self, message=f'Select new value for {name}',
                                               caption='Select Value', choices=options)
                    if dia.ShowModal()==wx.ID_OK and dia.GetSelection()>=0:
                        value = options[dia.GetSelection()]
                        for key in self.activated_leaf[1:-1]:
                            prev_dict = prev_dict[key]
                        prev_dict[self.activated_leaf[-1]] = value
                        self.tree.SetItemData(event.GetItem(), (name, f'{value} ({type(None).__name__})',
                                                                self.activated_leaf, type(None)))
                        self.show_item(event)
                        debug(f'updated {name} to {prev_dict[self.activated_leaf[-1]]}')
                        self.update_parents(event.GetItem(), self.activated_leaf)
                    return
            if vtype is list:
                for key in self.activated_leaf[1:-1]:
                    prev_dict = prev_dict[key]
                prev_value = prev_dict[self.activated_leaf[-1]]
                options = [str(di) for di in prev_value]
                dia = wx.SingleChoiceDialog(self, message=f'Select new value to edit in {name}',
                                            caption='Select Value', choices=options)
                if dia.ShowModal()==wx.ID_OK and dia.GetSelection()>=0:
                    idx = dia.GetSelection()
                    dia = wx.TextEntryDialog(self, message=f'Enter new value',
                                             caption='Enter Value',
                                             value=str(prev_value[idx]))
                    if dia.ShowModal()==wx.ID_OK:
                        if type(prev_value[idx]) is dict:
                            prev_value[idx] = eval(dia.GetValue())
                        else:
                            prev_value[idx] = type(prev_value[idx])(dia.GetValue())
                    self.tree.SetItemData(event.GetItem(), (name, f'{prev_value} (list)', self.activated_leaf, list))
                    self.update_parents(event.GetItem(), self.activated_leaf)
                    self.text.Clear()
                    self.text.AppendText('%s:\n\n\t%s'%(name, f'{prev_value} (list)'))
                return
            elif vtype is datetime:
                #from wx.adv import TimePickerCtrl
                print('date')
                return
            elif not vtype in [str, int, float, None]:
                # can only edit simple leaf items that can be converted from a str up to now
                print(vtype)
                return
            for key in self.activated_leaf[1:-1]:
                prev_dict = prev_dict[key]
            prev_value = prev_dict[self.activated_leaf[-1]]
            dia = wx.TextEntryDialog(self, message=f'Enter new value for {name} with type {vtype}', caption='Enter Value',
                                     value=str(prev_value))
            if dia.ShowModal()==wx.ID_OK:
                if vtype is None:
                    value = dia.GetValue()
                    try:
                        value = eval(value)
                    except Exception:
                        pass
                    vtype = type(value)
                else:
                    value = vtype(dia.GetValue())
                prev_dict[self.activated_leaf[-1]] = value
                self.tree.SetItemData(event.GetItem(), (name, f'{value} ({vtype.__name__})',
                                                        self.activated_leaf, vtype))
                self.show_item(event)
                debug(f'updated {name} to {prev_dict[self.activated_leaf[-1]]}')
                self.update_parents(event.GetItem(), self.activated_leaf)

    def update_parents(self, item, path, initial=False):
        if len(path)==1:
            return
        if initial:
            parent = item
        else:
            parent = self.tree.GetItemParent(item)
        mpath = list(path[:-1])
        obj = self.orso_repr[mpath[0]]
        mdict = self.datasets[mpath.pop(0)].meta
        while len(mpath)>0:
            obj = getattr(obj, mpath[0], None)
            mdict = mdict[mpath.pop(0)]
        value = mdict
        self.tree.SetItemData(parent,
                              (path[-2], yaml.dump(value, indent=4, sort_keys=False
                                                   ).replace('    ', '\t').replace('\n', '\n\t'),
                               path[:-1], obj))
        self.update_parents(parent, path[:-1])

    def make_orso_conform(self, evt):
        Model.update_orso_meta(self.datasets)
        try:
            from orsopy import fileio
            self.orso_repr = [fileio.Orso(**di.meta) for di in self.datasets]
        except Exception:
            self.orso_repr = [None for di in self.datasets]
        self.tree.DeleteAllItems()
        self.leaf_ids = []
        self.build_tree(selected=0)

    def OnRightClick(self, event: wx.TreeEvent):
        item = event.GetItem()
        if item in self.leaf_ids:
            name, data, leaf, vtype = self.tree.GetItemData(event.GetItem())
            if vtype is list:
                entries = ['append item', 'move item']
                callback = self.list_popup_action
            else:
                entries = ['no action implemented']
                callback = lambda event: None
            popupmenu = wx.Menu()
            self._popup_tree_item = (item, leaf)
            self._popup_menu_ids = {}
            for entry in entries:
                menuItem = popupmenu.Append(-1, entry)
                self._popup_menu_ids[menuItem.GetId()] = entry
            popupmenu.Bind(wx.EVT_MENU, callback)
            self.tree.PopupMenu(popupmenu, event.GetPoint())
        else:
            try:
                key, text, path, obj = self.tree.GetItemData(item)
            except (TypeError, ValueError):
                return
            popupmenu = wx.Menu()
            if type(obj) is dict:
                entries = ['new key']
                entries += list(obj.keys())
            else:
                entries = list(obj.__annotations__.keys())
            self._popup_tree_item = (item, path)
            self._popup_menu_ids = {}
            for entry in entries:
                if type(obj) is dict:
                    if entry == 'new key':
                        text = entry
                    else:
                        text = f'delete {entry}'
                elif entry in obj._orso_optionals:
                    text = f"new {entry}*"
                else:
                    text = f"new {entry}"
                menuItem = popupmenu.Append(-1, text)
                self._popup_menu_ids[menuItem.GetId()] = entry
            popupmenu.Bind(wx.EVT_MENU, self.make_new_key)
            self.tree.PopupMenu(popupmenu, event.GetPoint())

    def resolve_type(self, ktype, node, key):
        from orsopy.fileio.base import get_args, get_origin
        from typing import Union
        if get_origin(ktype) is dict:
            return dict
        if get_origin(ktype) is list:
            return list
        if get_origin(ktype)==Union:
            args =  list(get_args(ktype))
            if args[-1] is type(None):
                # this is an optional parameter, if only one other type this is not actually a Union
                args.pop(-1)
            if len(args)==1:
                return self.resolve_type(args[0], node, key)
            else:
                options = [ci.__name__ for ci in args]
                dia = wx.SingleChoiceDialog(self,
                                    message=f'Select object type for {node}.{key}',
                                    caption='Select Item', choices=options)
                if dia.ShowModal()==wx.ID_OK and dia.GetSelection()>=0:
                    ktype = args[dia.GetSelection()]
                    return self.resolve_type(ktype, node, key)
                else:
                    return None
        return ktype

    def make_new_key(self, event: wx.CommandEvent):
        key, text, path, obj = self.tree.GetItemData(self._popup_tree_item[0])
        spath = list(path)
        mdict = self.datasets[spath.pop(0)].meta
        while len(spath)>0:
            mdict = mdict[spath.pop(0)]
        if type(obj) is dict:
            dkey = self._popup_menu_ids[event.GetId()]
            if dkey == 'new key':
                dia = wx.TextEntryDialog(self, message=f'Enter new key name for {key}', caption='Enter Key Name')
                if dia.ShowModal()==wx.ID_OK:
                    parent = self.tree.GetItemParent(self._popup_tree_item[0])
                    pkey, ptext, ppath, pobj = self.tree.GetItemData(parent)
                    from typing import Union, List
                    from orsopy.fileio.base import get_args, get_origin
                    try:
                        ktype = pobj.__annotations__[key]
                        if get_origin(ktype)==Union:
                            ktype = [arg for arg in get_args(ktype) if get_origin(arg)==dict][0]
                        ktype = get_args(ktype)[1]
                        ktype = self.resolve_type(ktype, key, 'item')
                    except (AttributeError, KeyError, IndexError):
                        ktype = None
                    if hasattr(ktype, '_orso_optionals'):
                        new =  ktype.empty()
                        for attr, atype in ktype.__annotations__.items():
                            if get_origin(atype)==Union:
                                atype = get_args(atype)[0]
                            if get_origin(atype) in [list, dict]:
                                continue
                            adia = wx.TextEntryDialog(self,
                                                     message=f'Enter value for attribute {attr} ({atype.__name__}).'
                                                             f'\nCancel for "None".',
                                                     caption='Enter Attribute Value')
                            if adia.ShowModal()==wx.ID_OK:
                                setattr(new, attr, atype(adia.GetValue()))
                        obj[dia.GetValue()] = new.to_dict()
                    else:
                        obj[dia.GetValue()] = None
            else:
                del(obj[dkey])
            self.tree.DeleteChildren(self._popup_tree_item[0])
            self.add_children(self._popup_tree_item[0], mdict, path, obj)
            self.update_parents(self._popup_tree_item[0], path+[key], initial=True)
            return
        ktype=obj.__annotations__[self._popup_menu_ids[event.GetId()]]
        ktype = self.resolve_type(ktype, key, self._popup_menu_ids[event.GetId()])
        if hasattr(ktype, '_orso_optionals'):
            new = ktype.empty()
            setattr(obj, self._popup_menu_ids[event.GetId()], new)
            mdict[self._popup_menu_ids[event.GetId()]] = new.to_dict()
        elif ktype is dict:
            new = {}
            setattr(obj, self._popup_menu_ids[event.GetId()], new)
            mdict[self._popup_menu_ids[event.GetId()]] = new
        elif ktype is list:
            setattr(obj, self._popup_menu_ids[event.GetId()], [])
            mdict[self._popup_menu_ids[event.GetId()]] = []
        else:
            setattr(obj, self._popup_menu_ids[event.GetId()], None)
            mdict[self._popup_menu_ids[event.GetId()]] = None
        self.tree.DeleteChildren(self._popup_tree_item[0])
        self.add_children(self._popup_tree_item[0], mdict, path, obj)
        self.update_parents(self._popup_tree_item[0], path+[key], initial=True)

    def list_popup_action(self, event: wx.CommandEvent):
        name, data, leaf, vtype = self.tree.GetItemData(self._popup_tree_item[0])
        action = self._popup_menu_ids[event.GetId()]
        if action == 'move item':
            pass
        elif action == 'append item':
            parent = self.tree.GetItemParent(self._popup_tree_item[0])
            pkey, ptext, ppath, pobj = self.tree.GetItemData(parent)
            prev_dict = self.datasets[leaf[0]].meta
            for key in leaf[1:-1]:
                prev_dict = prev_dict[key]
            lobj = prev_dict[leaf[-1]]
            from typing import Union
            from orsopy.fileio.base import get_args, get_origin
            try:
                ktype = pobj.__annotations__[name]
                if get_origin(ktype)==Union:
                    ktype = [arg for arg in get_args(ktype) if get_origin(arg)==list][0]
                ktype = get_args(ktype)[0]
            except (AttributeError, KeyError, IndexError):
                ktype = None
            ktype = self.resolve_type(ktype, name, 'item')
            if hasattr(ktype, '_orso_optionals'):
                new = ktype.empty()
                lobj.append(new.to_dict())
            else:
                dia = wx.TextEntryDialog(self, message=f'Enter value for new {name} item', caption='Enter Value')
                if dia.ShowModal()==wx.ID_OK:
                    lobj.append(dia.GetValue())
                else:
                    return
            self.tree.SetItemData(self._popup_tree_item[0], (name, f'{lobj} (list)', leaf, list))
            self.update_parents(self._popup_tree_item[0], leaf)
            self.text.Clear()
            self.text.AppendText('%s:\n\n\t%s'%(name, f'{lobj} (list)'))



"""
An interface to show output messages from python logging with convenient fielters.
"""

import wx
from typing import List

from .message_dialogs import ShowErrorDialog, ShowNotificationDialog
from ..model_actions import ModelAction, ActionHistory, ActionBlock


class ActionDisplayDialog(wx.Dialog):
    def __init__(self, parent, action: ModelAction):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE|wx.STAY_ON_TOP|wx.RESIZE_BORDER,
                           name=f'GenX action details')
        self.SetTitle(f'Action Details')
        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        vbox.Add(wx.StaticText(self, label=f'Name: {action.action_name}'))
        vbox.Add(wx.StaticText(self, label=f'\nDescription:'))
        msg=wx.TextCtrl(self, value=str(action), style=wx.TE_READONLY|wx.TE_MULTILINE|wx.TE_DONTWRAP)
        font=msg.GetFont()
        font.SetFamily(wx.FONTFAMILY_TELETYPE)
        msg.SetFont(font)
        vbox.Add(msg, 1, wx.EXPAND)

class HistoryDialog(wx.Dialog):
    actions: List[ModelAction]
    current_index: int=0
    changed_actions: ActionBlock

    def __init__(self, parent, history: ActionHistory):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE|wx.STAY_ON_TOP|wx.RESIZE_BORDER,
                           name='ActionHistory')
        self.SetTitle('GenX Action History')
        self.changed_actions=None
        self.history = history
        self.actions=history.undo_stack+list(reversed(history.redo_stack))
        self.current_index=len(history.undo_stack)
        self.build_layout()

        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.show_action_details, self.action_list)

        pos=parent.GetPosition()
        size=parent.GetSize()
        self.SetSize(400, int(size.height*0.6))
        self.SetPosition(wx.Point(pos.x+size.width//2, pos.y+int(size.height*0.2)))

    def build_layout(self):
        vbox=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)

        self.action_list=wx.ListCtrl(self, style=wx.LC_REPORT)
        self.action_list.AppendColumn('Action')
        self.append_actions()

        self.resize_columns()
        vbox.Add(self.action_list, 1, wx.EXPAND)

        hbox=wx.BoxSizer(wx.HORIZONTAL)
        vbox.Add(hbox)
        cbutton=wx.Button(self, -1, label='Close')
        hbox.Add(cbutton)
        self.Bind(wx.EVT_BUTTON, self.OnDestroy, cbutton)

        ubutton=wx.Button(self, -1, label='Revert selected actions (and remove from history)')
        self.Bind(wx.EVT_BUTTON, self.OnRevert, ubutton)
        hbox.Add(ubutton)

    def OnDestroy(self, evt):
        self.EndModal(wx.ID_CANCEL)

    def OnRevert(self, evt):
        start=self.action_list.GetFirstSelected()
        if 0<=start<self.current_index:
            success = True
            items=[[start, 1]]
            nitem=self.action_list.GetNextSelected(start)
            while 0<=nitem<=self.current_index:
                if nitem==sum(items[-1]):
                    items[-1][1]+=1
                else:
                    items.append([nitem, 1])
                nitem = self.action_list.GetNextSelected(nitem)
            changed: List[ModelAction]=[]
            for start, length in items:
                try:
                    changed+=self.history.remove_actions(self.current_index-start, length).actions
                except Exception as e:
                    ShowErrorDialog(self, f'The actions could not be re-applied:\n{e}\n\n'
                                          f'The history was reset to the previous state. '
                                          f'You might need to analyze your stepts in detail to revert these actions')
                    self.actions = self.history.undo_stack+list(reversed(self.history.redo_stack))
                    self.current_index = len(self.history.undo_stack)
                    changed+=self.history.redo_stack[-length:]
                    self.action_list.DeleteAllItems()
                    self.append_actions()
                    success=False
                    break
            self.changed_actions=ActionBlock(self.actions[-1].model, changed)
            if success:
                self.EndModal(wx.ID_OK)
        else:
            ShowNotificationDialog(self, 'You have to select at least one action to be removed from the undo stack')

    def append_actions(self):
        for i, action in enumerate(self.actions):
            self.action_list.Append((action.action_name, ))
            if i>=self.current_index:
                self.action_list.SetItemTextColour(i, wx.Colour(100,100,100))

    def show_action_details(self, event):
        action=self.actions[event.GetIndex()]
        ActionDisplayDialog(self, action).ShowModal()

    def resize_columns(self):
        self.action_list.SetColumnWidth(0, wx.LIST_AUTOSIZE)
        self.action_list.EnsureVisible(self.action_list.GetItemCount()-1)



"""
Test that all functions implemented in Numba yield same results as pure python functions.
"""
import unittest
import numpy as np

from genx.models import lib, sxrd


lib.USE_NUMBA = False
from genx.models.lib import instrument, instrument_numba, paratt, paratt_numba, neutron_refl, neutron_numba, \
    surface_scattering


try:
    from genx.models.lib import paratt_cuda, neutron_cuda
except:
    CUDA = False
else:
    CUDA = True


class TestInstrumentModule(unittest.TestCase):

    # Test the models.core.instruments functions implemented in models.core.instruments_numba

    def test_gauss_symmetric(self):
        x = np.linspace(0., 5., 1000)
        s12 = 10.0  # sample length/2.
        sigma = 0.5  # beam width sigma
        G1 = instrument.GaussIntensity(x, s12, s12, sigma)
        G2 = instrument_numba.GaussIntensity(x, s12, s12, sigma)
        np.testing.assert_array_almost_equal(G1, G2)

    def test_gauss_nonsymmetric(self):
        x = np.linspace(0., 5., 1000)
        s1 = 10.0  # sample length left
        s2 = 5.0  # sample length right
        sigma = 0.5  # beam width sigma
        G1 = instrument.GaussIntensity(x, s1, s2, sigma)
        G2 = instrument_numba.GaussIntensity(x, s1, s2, sigma)
        np.testing.assert_array_almost_equal(G1, G2)

    def test_square(self):
        x = np.linspace(0., 5., 1000)
        slen = 10.0  # sample length
        beamwidth = 0.5  # beam width sigma
        S1 = instrument.SquareIntensity(x, slen, beamwidth)
        S2 = instrument_numba.SquareIntensity(x, slen, beamwidth)
        np.testing.assert_array_almost_equal(S1, S2)

    def test_tth2q_scalar(self):
        x = np.linspace(0., 0.5, 1000)
        wavelength = 4.5
        Q1 = instrument.TwoThetatoQ(wavelength, x)
        Q2 = instrument_numba.TwoThetatoQ(wavelength, x)
        np.testing.assert_array_almost_equal(Q1, Q2)

    def test_tth2q_vector(self):
        x = np.linspace(0., 0.5, 1000)
        wavelength = x*0.001+4.5
        Q1 = instrument.TwoThetatoQ(wavelength, x)
        Q2 = instrument_numba.TwoThetatoQ(wavelength, x)
        np.testing.assert_array_almost_equal(Q1, Q2)

    def test_q2th_scalar(self):
        x = np.linspace(0., 0.5, 1000)
        wavelength = 4.5
        TH1 = instrument.QtoTheta(wavelength, x)
        TH2 = instrument_numba.QtoTheta(wavelength, x)
        np.testing.assert_array_almost_equal(TH1, TH2)

    def test_q2th_vector(self):
        x = np.linspace(0., 0.5, 1000)
        wavelength = x*0.001+4.5
        TH1 = instrument.QtoTheta(wavelength, x)
        TH2 = instrument_numba.QtoTheta(wavelength, x)
        np.testing.assert_array_almost_equal(TH1, TH2)

    def test_resolutionvector_scalar(self):
        x = np.linspace(0., 0.5, 1000)
        dx = 0.01
        points = 10
        range = 3.5
        Q1, weight1 = instrument.ResolutionVector(x, dx, points, range)
        Q2, weight2 = instrument_numba.ResolutionVector(x, dx, points, range)
        np.testing.assert_array_almost_equal(Q1, Q2)
        np.testing.assert_array_almost_equal(weight1, weight2)

    def test_resolutionvector_vector(self):
        x = np.linspace(0., 0.5, 1000)
        dx = np.maximum(x*0.001, 1e-8)
        points = 10
        range = 3.5
        Q1, weight1 = instrument.ResolutionVector(x, dx, points, range)
        Q2, weight2 = instrument_numba.ResolutionVector(x, dx, points, range)
        np.testing.assert_array_almost_equal(Q1, Q2)
        np.testing.assert_array_almost_equal(weight1, weight2)


class TestParattModule(unittest.TestCase):

    # Test the models.core.paratt functions implemented in models.core.paratt_numba

    def test_refl_int_no_roughness(self):
        theta = np.linspace(0., 5., 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=np.float64)
        G1 = paratt.Refl(theta, lamda, n, d, sigma, return_int=True)
        G2 = paratt_numba.Refl(theta, lamda, n, d, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.Refl(theta, lamda, n, d, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_int_roughness(self):
        theta = np.linspace(0., 5., 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([10, 10, 10, 10, 10, 10, 10, 10], dtype=np.float64)
        G1 = paratt.Refl(theta, lamda, n, d, sigma, return_int=True)
        G2 = paratt_numba.Refl(theta, lamda, n, d, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.Refl(theta, lamda, n, d, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_amp_no_roughness(self):
        theta = np.linspace(0., 5., 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=np.float64)
        G1 = paratt.Refl(theta, lamda, n, d, sigma, return_int=False)
        G2 = paratt_numba.Refl(theta, lamda, n, d, sigma, return_int=False)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.Refl(theta, lamda, n, d, sigma, return_int=False)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_amp_roughness(self):
        theta = np.linspace(0., 5., 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([10, 10, 10, 10, 10, 10, 10, 10], dtype=np.float64)
        G1 = paratt.Refl(theta, lamda, n, d, sigma, return_int=False)
        G2 = paratt_numba.Refl(theta, lamda, n, d, sigma, return_int=False)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.Refl(theta, lamda, n, d, sigma, return_int=False)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_reflq_int_no_roughness(self):
        Q = np.linspace(0., 0.5, 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=np.float64)
        G1 = paratt.ReflQ(Q, lamda, n, d, sigma, return_int=True)
        G2 = paratt_numba.ReflQ(Q, lamda, n, d, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.ReflQ(Q, lamda, n, d, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_reflq_int_roughness(self):
        Q = np.linspace(0., 0.5, 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([10, 10, 10, 10, 10, 10, 10, 10], dtype=np.float64)
        G1 = paratt.ReflQ(Q, lamda, n, d, sigma, return_int=True)
        G2 = paratt_numba.ReflQ(Q, lamda, n, d, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.ReflQ(Q, lamda, n, d, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_reflq_amp_no_roughness(self):
        Q = np.linspace(0., 0.5, 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=np.float64)
        G1 = paratt.ReflQ(Q, lamda, n, d, sigma, return_int=False)
        G2 = paratt_numba.ReflQ(Q, lamda, n, d, sigma, return_int=False)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.ReflQ(Q, lamda, n, d, sigma, return_int=False)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_reflq_amp_roughness(self):
        Q = np.linspace(0., 0.5, 1000, dtype=np.float64)
        lamda = 4.5
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([10, 10, 10, 10, 10, 10, 10, 10], dtype=np.float64)
        G1 = paratt.ReflQ(Q, lamda, n, d, sigma, return_int=False)
        G2 = paratt_numba.ReflQ(Q, lamda, n, d, sigma, return_int=False)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.ReflQ(Q, lamda, n, d, sigma, return_int=False)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_nvary_int_roughness(self):
        theta = np.linspace(0., 5., 1000, dtype=np.float64)
        lamda = np.linspace(4., 5.0, 1000, dtype=np.float64)
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        n = n[:, np.newaxis]*lamda[np.newaxis, :]/4.0
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([10, 10, 10, 10, 10, 10, 10, 10], dtype=np.float64)
        G1 = paratt.Refl_nvary2(theta, lamda, n, d, sigma, return_int=True)
        G2 = paratt_numba.Refl_nvary2(theta, lamda, n, d, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.Refl_nvary2(theta, lamda, n, d, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_nvary_amp_roughness(self):
        theta = np.linspace(0., 5., 1000, dtype=np.float64)
        lamda = np.linspace(4., 5.0, 1000, dtype=np.float64)
        n = np.array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j,
                      1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j, 1], dtype=np.complex128)
        n = n[:, np.newaxis]*lamda[np.newaxis, :]/4.0
        d = np.array([2, 80, 20, 80, 20, 80, 20, 2], dtype=np.float64)
        sigma = np.array([10, 10, 10, 10, 10, 10, 10, 10], dtype=np.float64)
        G1 = paratt.Refl_nvary2(theta, lamda, n, d, sigma, return_int=False)
        G2 = paratt_numba.Refl_nvary2(theta, lamda, n, d, sigma, return_int=False)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = paratt_cuda.Refl_nvary2(theta, lamda, n, d, sigma, return_int=False)
            np.testing.assert_array_almost_equal(G1, G2)


class TestNeutronModule(unittest.TestCase):

    # Test the models.core.neutron_refl functions implemented in models.core.neutron_numba

    def test_refl_roughness(self):
        Q = np.linspace(0.001, 0.5, 1000, dtype=np.float64)
        # sld_Fe=8e-6
        sld_Fe_p = 12.9e-6
        sld_Fe_m = 2.9e-6
        sld_Pt = 6.22e-6

        Vp = np.array([sld_Pt/np.pi, sld_Fe_p/np.pi, sld_Pt/np.pi, sld_Fe_p/np.pi, 0], dtype=np.complex128)
        Vm = np.array([sld_Pt/np.pi, sld_Fe_m/np.pi, sld_Pt/np.pi, sld_Fe_m/np.pi, 0], dtype=np.complex128)
        d = np.array([3, 100, 50, 100, 3], dtype=np.float64)
        M_ang = np.array([0.0, 45*np.pi/180, 0.0, 90*np.pi/180, 0.0, ], dtype=np.float64)
        sigma = np.array([10., 10., 10., 10., 10.0], dtype=np.float64)
        G1 = neutron_refl.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        G2 = neutron_numba.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = neutron_cuda.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_no_roughness(self):
        Q = np.linspace(0.001, 0.5, 1000, dtype=np.float64)
        # sld_Fe=8e-6
        sld_Fe_p = 12.9e-6
        sld_Fe_m = 2.9e-6
        sld_Pt = 6.22e-6

        Vp = np.array([sld_Pt/np.pi, sld_Fe_p/np.pi, sld_Pt/np.pi, sld_Fe_p/np.pi, 0], dtype=np.complex128)
        Vm = np.array([sld_Pt/np.pi, sld_Fe_m/np.pi, sld_Pt/np.pi, sld_Fe_m/np.pi, 0], dtype=np.complex128)
        d = np.array([3, 100, 50, 100, 3], dtype=np.float64)
        M_ang = np.array([0.0, 45*np.pi/180, 0.0, 90*np.pi/180, 0.0, ], dtype=np.float64)
        sigma = None
        G1 = neutron_refl.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        G2 = neutron_numba.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = neutron_cuda.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_nonmag(self):
        # test ambient layer non-air
        Q = np.linspace(0.001, 0.5, 1000, dtype=np.float64)
        SLDs = np.array([6.36e+00, (4.66e+00-1.60e-02j), 2.07e+00], dtype=np.complex128)*1e-6
        n = 1.0-SLDs
        n_prime = 1.0-(SLDs-SLDs[-1])  # use corrected SLD for matrix method

        Vp = ((2.0*np.pi/4.5)**2*(1-n_prime**2)).astype(np.complex128)
        Vm = Vp
        d = np.array([0, 100, 0], dtype=np.float64)
        M_ang = np.array([0.0, 0.0, 0.0, ], dtype=np.float64)
        sigma = np.array([10., 3., 0.], dtype=np.float64)

        G0 = paratt.ReflQ(Q, 4.5, n, d, sigma, return_int=True)
        G1 = neutron_refl.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        G2 = neutron_numba.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G0, G1[0], decimal=5)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = neutron_cuda.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_nonmag_no_roughness(self):
        # test ambient layer non-air
        Q = np.linspace(0.001, 0.5, 1000, dtype=np.float64)
        SLDs = np.array([6.36e+00, (4.66e+00-1.60e-02j), 2.07e+00], dtype=np.complex128)*1e-6
        n = 1.0-SLDs
        n_prime = 1.0-(SLDs-SLDs[-1])  # use corrected SLD for matrix method

        Vp = ((2.0*np.pi/4.5)**2*(1-n_prime**2)).astype(np.complex128)
        Vm = Vp

        d = np.array([0, 100, 0], dtype=np.float64)
        M_ang = np.array([0.0, 0.0, 0.0, ], dtype=np.float64)
        sigma = None

        G0 = paratt.ReflQ(Q, 4.5, n, d, d*0., return_int=True)
        G1 = neutron_refl.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        G2 = neutron_numba.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G0, G1[0], decimal=4)
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = neutron_cuda.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_air_nonmag(self):
        # test ambient layer non-air
        Q = np.linspace(0.001, 0.5, 1000, dtype=np.float64)
        SLDs = np.array([6.36e+00, (4.66e+00-1.60e-02j), 0e+00], dtype=np.complex128)*1e-6
        n = 1.0-SLDs

        Vp = ((2.0*np.pi/4.5)**2*(1-n**2)).astype(np.complex128)
        Vm = Vp
        d = np.array([0, 100, 0], dtype=np.float64)
        M_ang = np.array([0.0, 0.0, 0.0, ], dtype=np.float64)
        sigma = np.array([10., 3., 0.], dtype=np.float64)

        G0 = paratt.ReflQ(Q, 4.5, n, d, sigma, return_int=True)
        G1 = neutron_refl.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        G2 = neutron_numba.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G0, G1[0])
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = neutron_cuda.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)

    def test_refl_nonmag_air_no_roughness(self):
        # test ambient layer non-air
        Q = np.linspace(0.001, 0.5, 1000, dtype=np.float64)
        SLDs = np.array([6.36e+00, (4.66e+00-1.60e-02j), 0e+00], dtype=np.complex128)*1e-6
        n = 1.0-SLDs

        Vp = ((2.0*np.pi/4.5)**2*(2*SLDs-SLDs**2)).astype(
            np.complex128)  # ((2.0*np.pi/4.5)**2*(1-n**2)).astype(np.complex128)
        Vm = Vp

        d = np.array([0, 100, 0], dtype=np.float64)
        M_ang = np.array([0.0, 0.0, 0.0, ], dtype=np.float64)
        sigma = None

        G0 = paratt.ReflQ(Q, 4.5, n, d, d*0., return_int=True)
        G1 = neutron_refl.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        G2 = neutron_numba.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
        np.testing.assert_array_almost_equal(G0, G1[0])  # Matrix not so accurate without magnetization
        np.testing.assert_array_almost_equal(G1, G2)
        if CUDA:
            G2 = neutron_cuda.Refl(Q, Vp, Vm, d, M_ang, sigma, return_int=True)
            np.testing.assert_array_almost_equal(G1, G2)


class TestSXRD(unittest.TestCase):
    # test the models.lib.surface_scattering fast complex sum implementation

    def setUp(self):
        sxrd.surface_lattice_sum = surface_scattering.surface_lattice_sum

    def test_calc_f(self):
        unitcell = sxrd.UnitCell(3.9045, 3.9045, 3.9045, 90, 90, 90)

        inst = sxrd.Instrument(wavel=1.0, alpha=1.0)

        bulk = sxrd.Slab()
        bulk.add_atom('Sr', 'sr2p', 0.0, 0.0, 0.0, 0.08, 1.0)
        bulk.add_atom('Ti', 'ti4p', 0.5, 0.5, 0.5, 0.08, 1.0)
        bulk.add_atom('O1', 'o2m', 0.5, 0.5, 0.0, 0.08, 1.0)
        bulk.add_atom('O2', 'o2m', 0.5, 0.0, 0.5, 0.08, 1.0)
        bulk.add_atom('O3', 'o2m', 0.0, 0.5, 0.5, 0.08, 1.0)

        stouc = sxrd.Slab()
        stouc.add_atom('Sr', 'sr2p', 0.0, 0.0, 0.0, 0.08, 1.0, 1)
        stouc.add_atom('Ti', 'ti4p', 0.5, 0.5, 0.5, 0.08, 1.0, 1)
        stouc.add_atom('O1', 'o2m', 0.5, 0.5, 0.0, 0.08, 1.0, 1)
        stouc.add_atom('O2', 'o2m', 0.0, 0.5, 0.5, 0.08, 1.0, 2)

        laouc = sxrd.Slab(c=1.05)
        laouc.add_atom('La', 'la3p', 0.0, 0.0, 0.0, 0.08, 1.0, 1)
        laouc.add_atom('Al', 'al3p', 0.5, 0.5, 0.5, 0.08, 1.0, 1)
        laouc.add_atom('O1', 'o2m', 0.5, 0.5, 0.0, 0.08, 1.0, 1)
        laouc.add_atom('O2', 'o2m', 0.5, 0.0, 0.5, 0.08, 1.0, 2)

        lao_interface = laouc.copy()
        lao_surface = laouc.copy()
        p4 = [sxrd.SymTrans([[1, 0], [0, 1]]), sxrd.SymTrans([[-1, 0], [0, -1]]),
              sxrd.SymTrans([[0, -1], [1, 0]]), sxrd.SymTrans([[0, 1], [-1, 0]]),
              sxrd.SymTrans([[-1, 0], [0, 1]]), sxrd.SymTrans([[1, 0], [0, -1]]),
              sxrd.SymTrans([[0, 1], [1, 0]]), sxrd.SymTrans([[0, -1], [-1, 0]])]
        sample = sxrd.Sample(inst, bulk, [stouc]+[lao_interface]+[laouc]*3+[lao_surface], unitcell)
        sample.set_surface_sym(p4)

        l = np.linspace(0., 5., 1000)
        o = l*0

        for h in range(5):
            for k in range(5):
                with self.subTest(f'h={h}, k={k}'):
                    res = sample.calc_f(o+h, o+k, l)
                    res_nb = sample.turbo_calc_f(o+h, o+k, l)
                    np.testing.assert_array_almost_equal(res, res_nb)


if __name__=='__main__':
    unittest.main()
